"""WHAD Protocol Logitech Unifying domain message abstraction layer.
"""
from typing import List, Union
from dataclasses import dataclass, field, fields

from whad.hub.registry import Registry
from whad.hub.message import HubMessage, pb_bind
from whad.hub import ProtocolHub
from whad.hub.esb import EsbNodeAddress
from whad.hub.metadata import Metadata
from whad.scapy.layers.esb import ESB_Hdr, ESB_Payload_Hdr, ESB_Ack_Response, ESB_Pseudo_Packet


class Commands:
    """Unifying commands
    """
    SetNodeAddress = 0x00
    Sniff = 0x01
    Jam = 0x02
    Send = 0x03
    SendRaw = 0x04
    LogitechDongleMode = 0x05
    LogitechKeyboardMode = 0x06
    LogitechMouseMode = 0x07
    Start = 0x08
    Stop = 0x09
    SniffPairing = 0x0a


@dataclass(repr=False)
class UnifyingMetadata(Metadata):
    is_crc_valid : bool = None
    address : str = None

    def convert_to_header(self):
        return None, self.timestamp

    @classmethod
    def convert_from_header(cls, pkt):
        metadata = UnifyingMetadata()
        pkt = ESB_Hdr(bytes(pkt))
        metadata.address = EsbNodeAddress(pkt.address)
        metadata.is_crc_valid = pkt.valid_crc
        metadata.timestamp = int(100000 * pkt.time)
        metadata.channel = 0
        return metadata

def generate_unifying_metadata(message, msg_type):
    metadata = UnifyingMetadata()

    if msg_type == "raw_pdu":
        message = message.raw_pdu
    elif msg_type == "pdu":
        message = message.pdu

    if message.HasField("rssi"):
        metadata.rssi = message.rssi
    metadata.channel = message.channel
    if message.HasField("timestamp"):
        metadata.timestamp = message.timestamp
    if message.HasField("crc_validity"):
        metadata.is_crc_valid = message.crc_validity
    if message.HasField("address"):
        metadata.address = ":".join(["{:02x}".format(i) for i in message.address])
    return metadata

@pb_bind(ProtocolHub, name="unifying", version=1)
class UnifyingDomain(Registry):
    """WHAD Logitech Unifying domain messages parser/factory.
    """

    NAME = 'unifying'
    VERSIONS = {}

    def __init__(self, version: int):
        """Initializes a Logitech Unifying domain instance
        """
        self.proto_version = version
        from whad.scapy.layers.unifying import bind
        bind()

    @staticmethod
    def parse(proto_version: int, message) -> HubMessage:
        """Parses a WHAD ESB Domain message as seen by protobuf
        """
        message_type = message.unifying.WhichOneof('msg')
        message_clazz = UnifyingDomain.bound(message_type, proto_version)
        return message_clazz.parse(proto_version, message)

    def is_packet_compat(self, packet) -> bool:
        """Determine if a packet is an ESB packet.
        """
        return isinstance(packet.metadata, UnifyingMetadata)

    def convert_packet(self, packet) -> HubMessage:
        """Convert an ESB packet to SendPdu or SendBlePdu message.
        """
        if isinstance(packet.metadata, UnifyingMetadata):
            if packet.metadata.raw:
                return UnifyingDomain.bound('send_raw', self.proto_version).from_packet(
                    packet
                )
            else:
                return UnifyingDomain.bound('send', self.proto_version).from_packet(
                    packet
                )
        else:
            # Error
            return None


    def format(self, packet):
        """
        Converts a scapy packet with its metadata to a tuple containing a scapy packet with
        the appropriate header and the timestamp in microseconds.
        """
        if ESB_Hdr not in packet:
            if hasattr(packet, "metadata") and hasattr(packet.metadata, "address"):
                address = packet.metadata.address
            else:
                address = "11:22:33:44:55"
            packet = ESB_Hdr(address = address)/packet

        packet.preamble = 0xAA # force a rebuild
        formatted_packet = ESB_Pseudo_Packet(bytes(packet))

        timestamp = None
        if hasattr(packet, "metadata"):
            timestamp = packet.metadata.timestamp

        return formatted_packet, timestamp

    def create_set_node_address(self, node_address: EsbNodeAddress) -> HubMessage:
        """Create a SetNodeAddress message

        :param node_address: Node address to set (size must be 1-5 bytes)
        :type node_address: EsbNodeAddress
        :return: instance of `SetNodeAddress`
        """
        return UnifyingDomain.bound('set_node_addr', self.proto_version)(
            address=node_address.value
        )

    def create_start(self) -> HubMessage:
        """Create a Start message

        :return: instance of `Start` message
        """
        return UnifyingDomain.bound('start', self.proto_version)()


    def create_stop(self) -> HubMessage:
        """Create a Stop message

        :return: instance of `Stop` message
        """
        return UnifyingDomain.bound('stop', self.proto_version)()


    def create_jam_mode(self, channel: int) -> HubMessage:
        """Create a JamMode message

        :param channel: ESB channel to jam
        :type channel: int
        :return: instance of `JamMode`
        """
        return UnifyingDomain.bound('jam', self.proto_version)(
            channel=channel
        )

    def create_sniff_mode(self, address: EsbNodeAddress, channel: int = 0xFF, show_acks: bool = False) -> HubMessage:
        """Create a SniffMode message

        :param address: Node address to filter
        :type address: EsbNodeAddress
        :param channel: Channel to sniff
        :type channel: int, optional
        :param show_acks: show acknowledgements
        :type show_acks: bool, optional
        :return: instance of `SniffMode`
        """
        return UnifyingDomain.bound('sniff', self.proto_version)(
            address=address.value,
            channel=channel,
            show_acks=show_acks
        )

    def create_jammed(self, timestamp: int):
        """Create a Jammed notification message

        :param timestamp: Timestamp at which the jamming has succeeded
        :type timestamp: int
        :return: instance of `Jammed`
        """
        return UnifyingDomain.bound('jammed', self.proto_version)(
            timestamp=timestamp
        )

    def create_dongle_mode(self, channel: int) -> HubMessage:
        """Create DongleMode message

        :param channel: Channel to listen on
        :type channel: int
        :return: instance of `DongleMode`
        """
        return UnifyingDomain.bound('dongle', self.proto_version)(
            channel=channel
        )

    def create_keyboard_mode(self, channel: int) -> HubMessage:
        """Create KeyboardMode message

        :param channel: Channel to listen on
        :type channel: int
        :return: instance of `KeyboardMode`
        """
        return UnifyingDomain.bound('keyboard', self.proto_version)(
            channel=channel
        )

    def create_mouse_mode(self, channel: int) -> HubMessage:
        """Create MouseMode message

        :param channel: Channel to listen on
        :type channel: int
        :return: instance of `MouseMode`
        """
        return UnifyingDomain.bound('mouse', self.proto_version)(
            channel=channel
        )

    def create_sniff_pairing(self) -> HubMessage:
        """Create SniffPairing message

        :return: instance of `SniffPairing`
        """
        return UnifyingDomain.bound('sniff_pairing', self.proto_version)()

    def create_send_pdu(self, channel: int, pdu: bytes, retr_count: int = 0):
        """Create a SendPdu message

        :param channel: Channel to use for transmission
        :type channel: int
        :param pdu: Data to send
        :type pdu: bytes
        :param retr_count: Retransmission count
        :type retr_count: int
        :return: instance of `SendPdu`
        """
        return UnifyingDomain.bound('send', self.proto_version)(
            channel=channel,
            pdu=pdu,
            retr_count=retr_count
        )

    def create_send_raw_pdu(self, channel: int, pdu: bytes, retr_count: int = 0):
        """Create a SendRawPdu message

        :param channel: Channel to use for transmission
        :type channel: int
        :param pdu: Raw data to send
        :type pdu: bytes
        :param retr_count: Retransmission count
        :type retr_count: int
        :return: instance of `SendPdu`
        """
        return UnifyingDomain.bound('send_raw', self.proto_version)(
            channel=channel,
            pdu=pdu,
            retr_count=retr_count
        )

    def create_pdu_received(self, channel: int, pdu: bytes, rssi: int = None, timestamp: int = None,
                          crc_validity: bool = None, address: EsbNodeAddress = None) -> HubMessage:
        """Create a PduReceived notification message.

        :param channel: Channel on which the PDU has been received
        :type channel: int
        :param pdu: Data received (PDU)
        :type pdu: bytes
        :param rssi: Received signal strength indicator
        :type rssi: int, optional
        :param timestamp: Reception timestamp
        :type timestamp: int, optional
        :param crc_validity: Indicate CRC validity
        :type crc_validity: bool, optional
        :param address: Sender address
        :type address: EsbNodeAddress, optional
        :return: instance of `PduReceived`
        """
        # Create our base message
        msg = UnifyingDomain.bound('pdu', self.proto_version)(
            channel=channel,
            pdu=pdu
        )

        # Add optional fields if provided
        if rssi is not None:
            msg.rssi = rssi
        if timestamp is not None:
            msg.timestamp = timestamp
        if address is not None:
            msg.address = address.value
        if crc_validity is not None:
            msg.crc_validity = crc_validity

        # Return message
        return msg


    def create_raw_pdu_received(self, channel: int, pdu: bytes, rssi: int = None, timestamp: int = None,
                          crc_validity: bool = None, address: EsbNodeAddress = None) -> HubMessage:
        """Create a RawPduReceived notification message.

        :param channel: Channel on which the PDU has been received
        :type channel: int
        :param pdu: Raw data received (PDU)
        :type pdu: bytes
        :param rssi: Received signal strength indicator
        :type rssi: int, optional
        :param timestamp: Reception timestamp
        :type timestamp: int, optional
        :param crc_validity: Indicate CRC validity
        :type crc_validity: bool, optional
        :param address: Sender address
        :type address: EsbNodeAddress, optional
        :return: instance of `PduReceived`
        """
        # Create our base message
        msg = UnifyingDomain.bound('raw_pdu', self.proto_version)(
            channel=channel,
            pdu=pdu
        )

        # Add optional fields if provided
        if rssi is not None:
            msg.rssi = rssi
        if timestamp is not None:
            msg.timestamp = timestamp
        if address is not None:
            msg.address = address.value
        if crc_validity is not None:
            msg.crc_validity = crc_validity

        # Return message
        return msg

from .address import SetNodeAddress
from .mode import UnifyingStart, UnifyingStop, JamMode, SniffMode, Jammed, DongleMode, \
    KeyboardMode, MouseMode, SniffPairing
from .pdu import PduReceived, RawPduReceived, SendPdu, SendRawPdu

__all__ = [
    'UnifyingDomain',
    'SetNodeAddress',
    'UnifyingStart',
    'UnifyingStop',
    'JamMode',
    'SniffMode',
    'Jammed',
    'DongleMode',
    'KeyboardMode',
    'MouseMode',
    'SniffPairing',
    'SendPdu',
    'SendRawPdu',
    'PduReceived',
    'RawPduReceived'
]



import logging
from whad.device import WhadDeviceConnector, Connector

logger = logging.getLogger(__name__)

class WhadMonitor:
    """
    Whad Monitor.

    A monitor is an object allowing to monitor the packets received and/or transmitted by a given
    connector. It can be attached to the target connector, and performs a specific action when a
    monitored packet is received or transmitted.
    """
    def __init__(self, monitor_reception=True, monitor_transmission=True):
        self._connector = None
        self.__monitor_reception = monitor_reception
        self.__monitor_transmission = monitor_transmission
        self.__setup_done = False
        self._processing = False

    def attach(self, connector):
        """
        Attachs the current monitor to a specific connector.

        :param connector: WhadConnector to monitor
        :return: boolean indicating the success of attach operation
        """
        success = False
        if isinstance(connector, (WhadDeviceConnector, Connector)):
            self._connector = connector
            if self.__monitor_reception:
                self._connector.attach_callback(self.process_packet, on_reception = True)
                #print("[i] reception callback attached.")
                logger.debug("monitor: reception callback attached.")
                success = True

            if self.__monitor_transmission:
                self._connector.attach_callback(self.process_packet, on_transmission = True)
                #print("[i] transmission callback attached.")
                logger.debug("monitor: transmission callback attached.")
                success = True
        return success

    def detach(self):
        """
        Detachs the current monitor from the previously attached connector.

        :return: boolean indicating the success of detach operation
        """
        logger.debug("monitor: detaching callbacks.")
        success = False
        if self._connector is not None:
            if self.__monitor_reception:
                self._connector.detach_callback(self.process_packet, on_reception = True)
                success = True
            if self.__monitor_transmission:
                self._connector.detach_callback(self.process_packet, on_transmission = True)
                success = True
        return success

    def setup(self):
        """
        Performs an action when the monitor is started for the first time (e.g., configuration).
        """
        pass

    def close(self):
        """
        Performs an action when the monitor is closed or destroyed.
        """
        pass

    def start(self):
        """
        Starts the monitor processing.
        """
        if not self.__setup_done:
            self.setup()
        self._processing = True

    def stop(self):
        """
        Stops the monitor processing.
        """
        self._processing = False

    def process_packet(self, packet):
        """
        Performs the monitoring action when a packet is received or transmitted by the targeted connector.

        :param packet: scapy packet to process
        """
        pass

    def __del__(self):
        self.close()

from .pcap import PcapWriterMonitor
from .wireshark import WiresharkMonitor



from whad.dot15d4.connector import Dot15d4, Sniffer, EndDevice, Coordinator
from whad.dot15d4.utils.phy import PHYS
__all__ = [
    'Dot15d4',
    'Sniffer',
    'Scanner',
    'EndDevice',
    'Coordinator',
    'PHYS'
]



"""RF4CE connectors
"""
from .base import RF4CE
from .sniffer import Sniffer
from .target import Target
from .controller import Controller

__all__ = [
    "RF4CE",
    "Sniffer",
    "Target",
    "Controller",
]


"""WHAD Protocol Logitech Unifying PDU messages abstraction layer.
"""
from whad.protocol.whad_pb2 import Message
from whad.scapy.layers.esb import ESB_Hdr, ESB_Payload_Hdr, ESB_Ack_Response, ESB_Pseudo_Packet
from ..message import pb_bind, PbFieldBytes, PbFieldBool, PbFieldInt, PbMessageWrapper, \
    dissect_failsafe
from whad.scapy.layers.unifying import bind
from whad.hub.message import AbstractPacket
from . import UnifyingDomain, UnifyingMetadata


@pb_bind(UnifyingDomain, 'send', 1)
class SendPdu(PbMessageWrapper):
    """ESB SendPdu message
    """

    channel = PbFieldInt('unifying.send.channel')
    pdu = PbFieldBytes('unifying.send.pdu')
    retr_count = PbFieldInt('unifying.send.retransmission_count')

    @dissect_failsafe
    def to_packet(self):
        """Convert SendPdu message to its scapy equivalent
        """
        return ESB_Payload_Hdr(bytes(self.pdu))

    @staticmethod
    def from_packet(packet, retr_count: int = 1):
        """Convert scapy packet to SendPdu message.
        """
        return SendPdu(
            channel=packet.metadata.channel,
            pdu=bytes(packet),
            retr_count=retr_count
        )

@pb_bind(UnifyingDomain, 'send_raw', 1)
class SendRawPdu(PbMessageWrapper):
    """ESB SendRawPdu message
    """

    channel = PbFieldInt('unifying.send_raw.channel')
    pdu = PbFieldBytes('unifying.send_raw.pdu')
    retr_count = PbFieldInt('unifying.send_raw.retransmission_count')

    @dissect_failsafe
    def to_packet(self):
        """Convert SendPdu message to its scapy equivalent
        """
        packet = ESB_Hdr(bytes(self.pdu))
        packet.preamble = 0xAA
        return packet


    @staticmethod
    def from_packet(packet, retr_count: int = 1):
        """Convert scapy packet to SendPdu message.
        """
        return SendRawPdu(
            channel=packet.metadata.channel,
            pdu=bytes(packet),
            retr_count=retr_count
        )

@pb_bind(UnifyingDomain, 'pdu', 1)
class PduReceived(PbMessageWrapper):
    """ESB PduReceived message
    """

    channel = PbFieldInt('unifying.pdu.channel')
    pdu = PbFieldBytes('unifying.pdu.pdu')
    rssi = PbFieldInt('unifying.pdu.rssi', optional=True)
    timestamp = PbFieldInt('unifying.pdu.timestamp', optional=True)
    crc_validity = PbFieldBool('unifying.pdu.crc_validity', optional=True)
    address = PbFieldBytes('unifying.pdu.address', optional=True)

    @dissect_failsafe
    def to_packet(self):
        """Convert message to the corresponding scapy packet
        """
        # Ensure our Unifying layers for scapy are correctly bound
        bind()

        # Build the Unifying packet (payload will be interpreted as Unifying header)
        packet = ESB_Payload_Hdr(bytes(self.pdu))
        packet.metadata = UnifyingMetadata()
        packet.metadata.channel = self.channel
        packet.metadata.raw = False

        packet.metadata.decrypted = False
        if self.rssi is not None:
            packet.metadata.rssi = self.rssi
        if self.timestamp is not None:
            packet.metadata.timestamp = self.timestamp
        if self.crc_validity is not None:
            packet.metadata.is_crc_valid = self.crc_validity
        if self.address is not None:
            packet.metadata.address = ":".join(["{:02x}".format(i) for i in self.address])
        return packet

    @staticmethod
    def from_packet(packet):
        """Convert scapy packet to PduReceived message
        """
        msg = PduReceived(
            channel=packet.metadata.channel,
            pdu=bytes(packet)
        )

        # Set message fields based on packet metadata
        if packet.metadata.decrypted is not None:
            msg.decrypted = packet.metadata.decrypted
        if packet.metadata.rssi is not None:
            msg.rssi = packet.metadata.rssi
        if packet.metadata.timestamp is not None:
            msg.timestamp = packet.metadata.timestamp
        if packet.metadata.is_crc_valid is not None:
            msg.crc_validity = packet.metadata.is_crc_valid
        if packet.metadata.address is not None:
            msg.address = bytes.fromhex(packet.metadata.address.replace(':', ''))

        return msg

@pb_bind(UnifyingDomain, 'raw_pdu', 1)
class RawPduReceived(PbMessageWrapper):
    """ESB RawPduReceived message
    """

    channel = PbFieldInt('unifying.raw_pdu.channel')
    pdu = PbFieldBytes('unifying.raw_pdu.pdu')
    rssi = PbFieldInt('unifying.raw_pdu.rssi', optional=True)
    timestamp = PbFieldInt('unifying.raw_pdu.timestamp', optional=True)
    crc_validity = PbFieldBool('unifying.raw_pdu.crc_validity', optional=True)
    address = PbFieldBytes('unifying.raw_pdu.address', optional=True)

    @dissect_failsafe
    def to_packet(self):
        """Convert message to the corresponding scapy packet
        """
        # Ensure our Unifying layers for scapy are correctly bound
        bind()

        # Build the Unifying packet (payload will be interpreted as Unifying header)
        packet = ESB_Hdr(bytes(self.pdu))
        packet.metadata = UnifyingMetadata()
        packet.metadata.channel = self.channel
        packet.metadata.raw = True

        packet.metadata.decrypted = False

        if self.rssi is not None:
            packet.metadata.rssi = self.rssi
        if self.timestamp is not None:
            packet.metadata.timestamp = self.timestamp
        if self.crc_validity is not None:
            packet.metadata.is_crc_valid = self.crc_validity
        if self.address is not None:
            packet.metadata.address = ":".join(["{:02x}".format(i) for i in self.address])
        return packet

    @staticmethod
    def from_packet(packet):
        """Convert scapy packet to RawPduReceived message
        """
        # Force packet preamble to 0xAA
        packet.preamble = 0xAA

        msg = RawPduReceived(
            channel=packet.metadata.channel,
            pdu=bytes(packet)
        )

        # Set message fields based on packet metadata

        if packet.metadata.decrypted is not None:
            msg.decrypted = packet.metadata.decrypted
        if packet.metadata.rssi is not None:
            msg.rssi = packet.metadata.rssi
        if packet.metadata.timestamp is not None:
            msg.timestamp = packet.metadata.timestamp
        if packet.metadata.is_crc_valid is not None:
            msg.crc_validity = packet.metadata.is_crc_valid
        if packet.metadata.address is not None:
            msg.address = bytes.fromhex(packet.metadata.address.replace(':', ''))

        return msg



"""WHAD Protocol BLE jamming messages abstraction layer.
"""
from whad.protocol.whad_pb2 import Message
from whad.hub.message import pb_bind, PbFieldInt, PbFieldBytes, PbMessageWrapper
from whad.hub.ble import BleDomain
from whad.hub.events import ConnectionEvt, DisconnectionEvt, DesyncEvt, SyncEvt

@pb_bind(BleDomain, 'connect', 1)
class ConnectTo(PbMessageWrapper):
    """BLE connect message class
    """
    bd_address = PbFieldBytes('ble.connect.bd_address')
    addr_type = PbFieldInt('ble.connect.addr_type')
    access_address = PbFieldInt('ble.connect.access_address', optional=True)
    channel_map = PbFieldBytes('ble.connect.channel_map', optional=True)
    hop_interval = PbFieldInt('ble.connect.hop_interval', optional=True)
    hop_increment = PbFieldInt('ble.connect.hop_increment', optional=True)
    crc_init = PbFieldInt('ble.connect.crc_init', optional=True)

@pb_bind(BleDomain, 'disconnect', 1)
class Disconnect(PbMessageWrapper):
    """BLE connect message class
    """
    conn_handle = PbFieldInt('ble.disconnect.conn_handle')

@pb_bind(BleDomain, 'synchronized', 1)
class Synchronized(PbMessageWrapper):
    """BLE synchronized notification message class.

    This class implements WHAD's AbstractEvent interface and can be converted
    into a `SyncEvt` object.
    """
    access_address = PbFieldInt('ble.synchronized.access_address')
    crc_init = PbFieldInt('ble.synchronized.crc_init')
    hop_interval = PbFieldInt('ble.synchronized.hop_interval')
    hop_increment = PbFieldInt('ble.synchronized.hop_increment')
    channel_map = PbFieldBytes('ble.synchronized.channel_map')

    def to_event(self):
        """Convert this message to an event.
        """
        return SyncEvt(
            access_address=self.access_address,
            crc_init=self.crc_init,
            hop_interval=self.hop_interval,
            hop_increment=self.hop_increment,
            channel_map=self.channel_map
        )

    @staticmethod
    def from_event(event):
        return Synchronized(
            access_address=event.access_address,
            crc_init=event.crc_init,
            hop_interval=event.hop_interval,
            hop_increment=event.hop_increment,
            channel_map=event.channel_map
        )

@pb_bind(BleDomain, 'connected', 1)
class Connected(PbMessageWrapper):
    """BLE connected message class.

    This class implements WHAD's AbstractEvent interface and can be converted
    into a `ConnectionEvt` object.
    """
    initiator = PbFieldBytes('ble.connected.initiator')
    advertiser = PbFieldBytes('ble.connected.advertiser')
    access_address = PbFieldInt('ble.connected.access_address')
    conn_handle = PbFieldInt('ble.connected.conn_handle')
    adv_addr_type = PbFieldInt('ble.connected.adv_addr_type')
    init_addr_type = PbFieldInt('ble.connected.init_addr_type')

    def to_event(self):
        """Convert this message to an event.
        """
        return ConnectionEvt(
            initiator=self.initiator,
            advertiser=self.advertiser,
            access_address=self.access_address,
            conn_handle=self.conn_handle,
            adv_addr_type=self.adv_addr_type,
            init_addr_type=self.init_addr_type
        )

    @staticmethod
    def from_event(event):
        return Connected(
            initiator=event.initiator,
            advertiser=event.advertiser,
            access_address=event.access_address,
            conn_handle=event.conn_handle,
            adv_addr_type=event.adv_addr_type,
            init_addr_type=event.init_addr_type
        )

@pb_bind(BleDomain, 'disconnected', 1)
class Disconnected(PbMessageWrapper):
    """BLE connected message class.

    This class implements WHAD's AbstractEvent interface and can be converted
    into a `DisconnectionEvt` object.
    """
    reason = PbFieldInt('ble.disconnected.reason')
    conn_handle = PbFieldBytes('ble.disconnected.conn_handle')

    def to_event(self):
        """Convert this message to an event.
        """
        return DisconnectionEvt(
            conn_handle=self.conn_handle,
            reason=self.reason
        )
    
    @staticmethod
    def from_event(event):
        return Disconnected(
            conn_handle=event.conn_handle,
            reason=event.reason
        )

@pb_bind(BleDomain, 'desynchronized', 1)
class Desynchronized(PbMessageWrapper):
    """BLE connect message class.

    This class implements WHAD's AbstractEvent interface and can be converted
    into a `DesyncEvt` object.
    """
    access_address = PbFieldInt('ble.desynchronized.access_address')

    def to_event(self):
        """Convert this message to an event.
        """
        return DesyncEvt(
            access_address=self.access_address,
        )
    
    @staticmethod
    def from_event(event):
        return Desynchronized(
            access_address=event.access_address,
        )



from struct import pack, unpack

class InformationAttributeValue:
    @classmethod
    def pack(cls, value):
        return value

    @classmethod
    def unpack(cls, value):
        return value

class PeripheralIDsValue(InformationAttributeValue):
    @classmethod
    def pack(cls, value):
        if not isinstance(value, list):
            raise ValueError("pack")

        packed = b""
        for id in value:
            if not isinstance(id, int):
                raise ValueError("pack")
            packed += pack("I", id)
        return packed

    @classmethod
    def unpack(cls, value):
        if not isinstance(value, bytes):
            raise ValueError("unpack")

        nb_elements = len(value) // 4
        return list(unpack(nb_elements * "I", value))

class SoftwareVersioningValue(InformationAttributeValue):
    @classmethod
    def pack(cls, major, minor, revision, patch):
        if (
            not isinstance(major, int) or major > 255 or major < 0 or
            not isinstance(minor, int) or minor > 255 or minor < 0 or
            not isinstance(revision, int) or revision > 255 or revision < 0 or
            not isinstance(patch, int) or patch > 255 or patch < 0
        ):
            raise ValueError("pack")

        return bytes([major, minor, revision, patch])

    @classmethod
    def unpack(cls, value):
        return unpack("BBBB", value)

class IRDBVersioningValue(SoftwareVersioningValue):
    pass

class HardwareVersioningValue(InformationAttributeValue):
    @classmethod
    def pack(cls, manufacturer, model, revision, lot_code):
        if (
            not isinstance(manufacturer, int) or manufacturer > 15 or manufacturer < 0 or
            not isinstance(model, int) or model > 15 or model < 0 or
            not isinstance(revision, int) or revision > 255 or revision < 0 or
            not isinstance(lot_code, int) or lot_code > 4095 or lot_code < 0
        ):
            raise ValueError("pack")

        return bytes(
            [
                ((manufacturer & 0xF) << 4) | (model & 0xF),
                revision,
                (lot_code >> 8),
                (lot_code & 0xFF)
            ]
        )

    @classmethod
    def unpack(cls, value):
        if not isinstance(value, bytes):
            raise ValueError("unpack")
        manufacturer, model = value[0] >> 4, value[0] & 0x0F
        revision = value[1]
        lot_value = unpack(">H", value[2:])[0]
        return manufacturer, model, revision, lot_value

class VersioningValue(InformationAttributeValue):
    @classmethod
    def pack(cls, software, hardware, irdb):
        return (
            SoftwareVersioningValue.pack(*software) +
            HardwareVersioningValue.pack(*hardware) +
            IRDBVersioningValue.pack(*irdb)
        )

    @classmethod
    def unpack(cls,value):
        return (
            SoftwareVersioningValue.unpack(value[:4]),
            HardwareVersioningValue.unpack(value[4:8]),
            IRDBVersioningValue.unpack(value[8:]),
        )

class FlagsBatteryValue(InformationAttributeValue):
    @classmethod
    def pack(cls, battery_replacement, battery_charging, impending_doom):
        if (
            not isinstance(battery_replacement, bool) or
            not isinstance(battery_charging, bool) or
            not isinstance(impending_doom, bool)
        ):
            raise ValueError("pack")
        return bytes(
            [
                (int(battery_replacement) & 1) |
                ((int(battery_charging) & 1) >> 1) |
                ((int(impending_doom) & 1) >> 2)
            ]
        )

    @classmethod
    def unpack(cls, value):
        if not isinstance(value, bytes):
            raise ValueError("unpack")

        battery_replacement = bool(value[0] & 0x1)
        battery_charging = bool((value[0] & 0x2) >> 1)
        impending_doom = bool((value[0] & 0x4) >> 2)
        return (battery_replacement, battery_charging, impending_doom)


class VoltageLevelValue(InformationAttributeValue):
    @classmethod
    def pack(cls, level):
        # linear interpolation between 0V & 4V

        if level < 0 or level > 4:
            raise ValueError("pack")

        return bytes([int((level / 4) * 255)])

    @classmethod
    def unpack(cls, value):
        if not isinstance(value, bytes):
            raise ValueError("unpack")

        return (value[0] / 255) * 4

class BatteryStatusValue(InformationAttributeValue):
    @classmethod
    def pack(cls, flags, loaded_voltage_level, number_of_rf_codes_transmitted, number_of_ir_codes_transmitted, unloaded_voltage_level):
        return (
            FlagsBatteryValue.pack(*flags) +
            VoltageLevelValue.pack(loaded_voltage_level) +
            pack("I", number_of_rf_codes_transmitted) +
            pack("I", number_of_ir_codes_transmitted) +
            VoltageLevelValue.pack(unloaded_voltage_level)
        )

    @classmethod
    def unpack(cls, value):
        if not isinstance(value, bytes):
            raise ValueError("unpack")

        return (
            FlagsBatteryValue.unpack(value[0]) ,
            VoltageLevelValue.unpack(value[1]),
            unpack("I", value[2:6]),
            VoltageLevelValue.unpack(value[11])
        )

class ValidationConfigurationValue(InformationAttributeValue):
    @classmethod
    def pack(cls, auto_check_validation_period, link_lost_wait_time):
        if not isinstance(auto_check_validation_period, int) or not isinstance(link_lost_wait_time, int):
            raise ValueError("pack")
        return pack("H", auto_check_validation_period) + pack("H", link_lost_wait_time)

    @classmethod
    def unpack(cls, value):
        if not isinstance(value, bytes):
            raise ValueError("unpack")

        return (
            unpack("H", value[:2])[0],
            unpack("H", value[2:])[0]
        )

class IRRFFlagsValue(InformationAttributeValue):
    @classmethod
    def pack(cls, rf_pressed_specified, rf_repeated_specified, rf_released_specified, ir_specified, use_default, permanent):
        if (
            not isinstance(rf_pressed_specified, bool) or
            not isinstance(rf_repeated_specified, bool) or
            not isinstance(rf_released_specified, bool) or
            not isinstance(ir_specified, bool) or
            not isinstance(permanent, bool) or
            not isinstance(use_default, bool)
        ):
            raise ValueError("pack")

        return bytes(
            [
                (int(rf_pressed_specified) & 0x1) |
                ((int(rf_repeated_specified) & 0x1) >> 1) |
                ((int(rf_released_specified) & 0x1) >> 2) |
                ((int(ir_specified) & 0x1) >> 3) |
                ((int(use_default) & 0x1) >> 6) |
                ((int(permanent) & 0x1) >> 7)
            ]
        )

    @classmethod
    def unpack(cls, value):
        if not isinstance(value, bytes):
            raise ValueError("unpack")

        return (
            value[0] & 0b00000001,
            (value[0] & 0b00000010) >> 1,
            (value[0] & 0b00000100) >> 2,
            (value[0] & 0b00001000) >> 3,
            (value[0] & 0b01000000) >> 6,
            (value[0] & 0b10000000) >> 7
        )

class RFConfigValue(InformationAttributeValue):
    @classmethod
    def pack(cls, minimum_number_of_transmissions, keep_transmitting_until_key_release, short_retry):
        if (
            not isinstance(minimum_number_of_transmissions, int) or
            not isinstance(keep_transmitting_until_key_release, bool) or
            not isinstance(short_retry, bool)
        ):
            raise ValueError("pack")

        return bytes([
            (minimum_number_of_transmissions & 0x0F) |
            ((int(keep_transmitting_until_key_release)  & 1) << 4) |
            ((int(short_retry)  & 1) << 5)
        ])

    @classmethod
    def unpack(cls, value):
        if not isinstance(value, bytes):
            raise ValueError("unpack")

        return (
            (minimum_number_of_transmissions & 0x0F),
            (keep_transmitting_until_key_release & 0x10) >> 4,
            (short_retry & 0x20) >> 5
        )

class IRConfigValue(InformationAttributeValue):
    @classmethod
    def pack(cls, minimum_number_of_transmissions, keep_transmitting_until_key_release, tweak_database):
        if (
            not isinstance(minimum_number_of_transmissions, int) or
            not isinstance(keep_transmitting_until_key_release, bool) or
            not isinstance(tweak_database, bool)
        ):
            raise ValueError("pack")

        return bytes([
            (minimum_number_of_transmissions & 0x0F) |
            ((int(keep_transmitting_until_key_release)  & 1) << 4) |
            ((int(tweak_database)  & 1) << 6)
        ])

    @classmethod
    def unpack(cls, value):
        if not isinstance(value, bytes):
            raise ValueError("unpack")

        return (
            (minimum_number_of_transmissions & 0x0F),
            (keep_transmitting_until_key_release & 0x10) >> 4,
            (tweak_database & 0x40) >> 6
        )

class TXOptionsValue(InformationAttributeValue):
    @classmethod
    def pack(cls, broadcast_transmission, extended_address_mode, acknowledgement_mode, security_enabled, channel_agility_mode, channel_normalization_mode, vendor_specific):
        if (
            not isinstance(broadcast_transmission, bool) or
            not isinstance(extended_address_mode, bool) or
            not isinstance(acknowledgement_mode, bool) or
            not isinstance(security_enabled, bool) or
            not isinstance(channel_agility_mode, bool) or
            not isinstance(channel_normalization_mode, bool) or
            not isinstance(vendor_specific, bool)
        ):
            raise ValueError("pack")

        return bytes(
            [
                (int(broadcast_transmission) & 0x1) |
                ((int(extended_address_mode) & 0x1) << 1) |
                ((int(acknowledgement_mode) & 0x1) << 2) |
                ((int(security_enabled) & 0x1) << 3) |
                ((int(channel_agility_mode) & 0x1) << 4) |
                ((int(channel_normalization_mode) & 0x1) << 5) |
                ((int(vendor_specific) & 0x1) << 6)
            ]
        )

    @classmethod
    def unpack(cls, value):
        if not isinstance(value, bytes):
            raise ValueError("unpack")

        return (
            (value[0] & 0b00000001),
            (value[0] & 0b00000010) >> 1,
            (value[0] & 0b00000100) >> 2,
            (value[0] & 0b00001000) >> 3,
            (value[0] & 0b00010000) >> 4,
            (value[0] & 0b00100000) >> 5,
            (value[0] & 0b01000000) >> 6
        )

class RFDescriptorValue(InformationAttributeValue):
    @classmethod
    def pack(cls, rf_config, tx_options, payload):
        if not isinstance(payload, bytes):
            raise ValueError("pack")

        return RFConfigValue.pack(*rf_config) + TXOptionsValue.pack(*tx_options) + bytes([len(payload)]) + payload

    @classmethod
    def unpack(cls, value):
        if not isinstance(value, bytes):
            raise ValueError("unpack")

        return (
            RFConfigValue.unpack(value[0]),
            TXOptionsValue.unpack(value[1]),
            value[2],
            value[3:]
        )


class IRDescriptorValue(InformationAttributeValue):
    @classmethod
    def pack(cls, ir_config, ir_code):
        if not isinstance(ir_code, bytes):
            raise ValueError("pack")

        return IRConfigValue.pack(*ir_config) + bytes([len(ir_code)]) + ir_code

    @classmethod
    def unpack(cls, value):
        if not isinstance(value, bytes):
            raise ValueError("unpack")

        return (
            IRConfigValue.unpack(value[0]),
            value[1],
            value[2:]
        )


class IRRFElementValue(InformationAttributeValue):
    @classmethod
    def pack(cls, flags, rf_pressed_descriptor, rf_repeated_descriptor, rf_released_descriptor, ir_descriptor):
        rf_pressed_specified = flags[0]
        rf_repeated_specified = flags[1]
        rf_released_specified = flags[2]
        ir_specified = flags[3]

        return (
            IRRFFlagsValue.pack(*flags) +
            (RFDescriptorValue.pack(*rf_pressed_descriptor) if rf_pressed_specified else b"") +
            (RFDescriptorValue.pack(*rf_repeated_descriptor) if rf_repeated_specified else b"") +
            (RFDescriptorValue.pack(*rf_released_descriptor) if rf_released_specified else b"") +
            (IRDescriptorValue.pack(*ir_descriptor) if ir_specified else b"")
        )

    @classmethod
    def unpack(cls, value):
        flags = IRRFFlagsValue.unpack(value[0])
        (
            rf_pressed_specified,
            rf_repeated_specified,
            rf_released_specified,
            ir_specified,
            use_default,
            permanent
        ) = flags

        value = value[1:]

        descriptors = []
        for i in (rf_pressed_specified, rf_repeated_specified, rf_released_descriptor):
             descriptor, value = RFDescriptorValue.unpack(value[:3 + value[2]]), value[3 + value[2]:]
             descriptors.append(descriptor)

        descriptors.append(IRDescriptorValue(value))
        return (flags, *descriptors)



"""
WHAD various utilities for transversal automation.
"""

from typing import Optional
from importlib import import_module
from inspect import getdoc
from dataclasses import fields, is_dataclass

from pkgutil import iter_modules
from scapy.config import conf

import whad

def gen_option_name(config_param_name: str) -> str:
    """Generate an option name from a parameter name.

    Basically, this function replaces all "_" with "-".

    >>> gen_option_name("spreading_factor")
    """
    return config_param_name.replace('_', '-')

def gen_config_param_name(option_name: str) -> str:
    """Generate a config parameter name from an option name.

    Basically, this function replaces all "-" with "_".
    """
    return option_name.replace('-', '_')

def list_implemented_sniffers():
    """Build a dictionnary of sniffers connector and configuration, by domain.
    """
    environment = {}

    # Iterate over modules
    for _, candidate_protocol,_ in iter_modules(whad.__path__):
        # If the module contains a sniffer connector and a sniffing module,
        # store the associated classes in the environment dictionary
        try:
            module = import_module("whad.{}.connector.sniffer".format(candidate_protocol))
            configuration_module = import_module("whad.{}.sniffing".format(candidate_protocol))
            environment[candidate_protocol] = {
                "sniffer_class":module.Sniffer,
                "configuration_class":configuration_module.SnifferConfiguration
            }
        except ModuleNotFoundError:
            pass
    # return the environment dictionary
    return environment

def get_sniffer_parameters(configuration_class):
    """
    Extract all parameters from a sniffer configuration class, with their name and associated documentation.

    :param configuration_class: sniffer configuration class
    :return: dict containing parameters for a given configuration class
    """
    parameters = {}
    # Extract documentation of every field in the configuration class
    fields_configuration_documentation = {
        i.replace(":param ","").split(":")[0] : i.replace(":param ","").split(":")[1]
        for i in getdoc(configuration_class).split("\n")
        if i.startswith(":param ")
    }

    # Iterate over the fields of the configuration class
    for field in fields(configuration_class):
        field_name = gen_option_name(field.name)

        # If the field is a dataclass, process subfields
        if is_dataclass(field.type):
            # Extract documentation of every subfields
            subfields_configuration_documentation = {
                i.replace(":param ","").split(":")[0] : i.replace(":param ","").split(":")[1]
                for i in getdoc(field.type).split("\n")
                if i.startswith(":param ")
            }

            # Populate parameters dict with subfields configuration
            for subfield in fields(field.type):
                subfield_name = gen_option_name(subfield.name)
                parameters["{}.{}".format(field.name,subfield.name)] = (
                    subfield.type,
                    subfield.default,
                    field.type,
                    (
                        subfields_configuration_documentation[subfield.name]
                        if subfield.name in subfields_configuration_documentation
                        else None
                    )
                )

        # if the field is not a dataclass, process it
        else:
            # Populate parameters dict with field configuration
            parameters[field_name] = (
                field.type,
                field.default,
                None,
                (
                    fields_configuration_documentation[field.name]
                    if field.name in fields_configuration_documentation
                    else None
                )
            )
    return parameters


def build_configuration_from_args(environment, args):
    """
    Build sniffer configuration from arguments provided via argparse.

    :param environment: environment
    :type environment: dict
    :param args: arguments provided by user
    :type args: :class:`argparse.ArgumentParser`
    """
    configuration = environment[args.domain]["configuration_class"]()
    subfields = {}
    for parameter in environment[args.domain]["parameters"]:

        parameter_real_field = gen_config_param_name(parameter)
        base_class = None

        base_class = environment[args.domain]["parameters"][parameter][2]
        if base_class is None:
            try:
                setattr(configuration,parameter_real_field, getattr(args,parameter))
            except AttributeError:
                pass
        else:
            main, sub = parameter.split(".")
            if main not in subfields:
                subfields[main] = base_class()
            sub_real_name = gen_config_param_name(sub)
            setattr(subfields[main], sub_real_name, getattr(args, parameter))

    for subfield in subfields:
        subfield_real_name = gen_config_param_name(subfield)
        create = False
        for item in fields(subfields[subfield]):
            item_real_name = gen_config_param_name(item.name)
            if getattr(subfields[subfield], item_real_name) is not None:
                create = True
                break
        if create:
            setattr(configuration, subfield_real_name, subfields[subfield])
        else:
            setattr(configuration, subfield_real_name, None)
    return configuration


def list_implemented_injectors() -> dict:
    """Build a dictionnary of injectors connector and configuration, by domain.
    """
    environment = {}

    # Iterate over modules
    for _, candidate_protocol,_ in iter_modules(whad.__path__):
        # If the module contains an injector connector and a injecting module,
        # store the associated classes in the environment dictionary
        try:
            module = import_module("whad.{}.connector.injector".format(candidate_protocol))
            configuration_module = import_module("whad.{}.injecting".format(candidate_protocol))
            environment[candidate_protocol] = {
                "injector_class":module.Injector,
                "configuration_class":configuration_module.InjectionConfiguration
            }
        except ModuleNotFoundError:
            pass
    # return the environment dictionary
    return environment


def get_injector_parameters(configuration_class):
    """
    Extract all parameters from a injector configuration class, with their name and associated documentation.

    :param configuration_class: injector configuration class
    :return: dict containing parameters for a given configuration class
    """
    parameters = {}
    # Extract documentation of every field in the configuration class
    fields_configuration_documentation = {
        i.replace(":param ","").split(":")[0] : i.replace(":param ","").split(":")[1]
        for i in getdoc(configuration_class).split("\n")
        if i.startswith(":param ")
    }

    # Iterate over the fields of the configuration class
    for field in fields(configuration_class):
        field_name = gen_option_name(field.name)

        # If the field is a dataclass, process subfields
        if is_dataclass(field.type):
            # Extract documentation of every subfields
            subfields_configuration_documentation = {
                i.replace(":param ","").split(":")[0] : i.replace(":param ","").split(":")[1]
                for i in getdoc(field.type).split("\n")
                if i.startswith(":param ")
            }

            # Populate parameters dict with subfields configuration
            for subfield in fields(field.type):
                subfield_name = gen_option_name(subfield.name)
                parameters["{}.{}".format(field.name,subfield.name)] = (
                    subfield.type,
                    subfield.default,
                    field.type,
                    (
                        subfields_configuration_documentation[subfield.name]
                        if subfield.name in subfields_configuration_documentation
                        else None
                    )
                )

        # if the field is not a dataclass, process it
        else:
            # Populate parameters dict with field configuration
            parameters[field_name] = (
                field.type,
                field.default,
                None,
                (
                    fields_configuration_documentation[field.name]
                    if field.name in fields_configuration_documentation
                    else None
                )
            )
    return parameters


def get_analyzers(protocol: Optional[str] = None) -> dict:
    """Retrieve every protocol analyzer available or a subset if a specific
    protocol name is provided.

    :param protocol: Consider only analyzers for the specified protocol.
    :type protocol: str
    :return: a dictionnary of protocol analyzers indexed by their protocol name.
    :rtype: dict
    """
    analyzers = {}
    for _, candidate_protocol,_ in iter_modules(whad.__path__):
        # If the module contains a list of analyzers,
        # store the associated analyzers in analyzers variable
        try:
            module = import_module("whad.{}.utils.analyzer".format(candidate_protocol))
            if candidate_protocol == protocol:
                analyzers = module.analyzers
                break
            elif protocol is None:
                analyzers[candidate_protocol] = module.analyzers
        except ModuleNotFoundError:
            pass
    return analyzers



"""PHY inspection tool

This utility provides specific tools for PHY inspection including:

- frequency scanning and RSSI correlation to determine on which channel a device transmits ;
- pattern searching including default bitstream transformation (BLE dewhitening, byte/bit swapping, endianness) ;
- similarity analysis that displays similar chunks of data with a specific color ;
- encryption guessing tools, based on packets analysis.

This tool is meant to be completely PHY agnostic and will work with any hardware that returns packets.

Guessed configuration can be saved for later use, in a dedicated configuration file.
"""
from time import sleep
from queue import Queue, Empty
from whad.cli.app import CommandLineApp, ApplicationError
from whad.phy.connector import Phy
from whad.ble.exceptions import PeripheralNotFound

import logging
logger = logging.getLogger(__name__)

#logging.basicConfig(level=logging.DEBUG)

class PhyCorrelator(Phy):

    def __init__(self, device, rssi=-100):
        super().__init__(device)
        self.__rssi = rssi
        self.__packet_queue = Queue()
        self.__callback = None

    def set_callback(self, callback):
        """Register a callback to be notified each time a packet
        is received.
        """
        self.__callback = callback

    def on_packet(self, packet):
        """On packet reception, call a callback if set or save the packet
        into our RX queue and wait for the queue to be read later.
        """
        if packet.metadata.rssi > self.__rssi:
            print(bytes(packet).hex())
            if self.__callback is not None:
                self.__callback(packet)
            else:
                self.__packet_queue.put(packet)

    def wait_packet(self, timeout=None):
        """Wait for a packet
        """
        try:
            packet = self.__pkt_queue.get(block=True, timeout=timeout)
            return packet
        except Empty as empty:
            return None

class FrequencyScanner(object):
    """Frequency scanner.

    This analysis module will loop on a range of frequencies and will only
    display packets with RSSI stronger than a given threshold. It can be used
    to easily guess the frequency used by a device to send packets by putting
    this device close to our receiver and make it send as much packets as
    possible.
    """

    def __init__(self, app : CommandLineApp, phy : PhyCorrelator, freq_start : int, freq_end : int,
                 freq_step : int, delay : int):
        """Initialize the frequency scanner.

        :param app: Command-line application this scanner is attached to
        :type app: CommandLineApp
        :param phy: PHY correlator to use
        :type phy: PhyCorrelator
        :param freq_start: Start frequency used for the frequency sweep, in Hz
        :type freq_start: int
        :param freq_end: End frequency used for the frequency sweep, in Hz
        :type freq_end: int
        :param freq_step: Frequency step value in Hz
        :type freq_step: int
        :param delay: Specifies the number of seconds to wait on each frequency
        :type delay: int
        """
        self.__app = app
        self.__phy = phy
        self.__delay = delay
        self.__frequencies = range(freq_start, freq_end, freq_step)
        self.__phy.set_callback(self.on_packet)
        self.__stats = {}
        self.__current_freq = 0
        for f in self.__frequencies:
            self.__stats[f] = 0

    def scan(self):
        """Loop on frequencies, wait `self.__delay` seconds on each and capture packets with RSSI
        bigger than a threshold.
        """
        print('Prepare your device and bring it close to the receiver hardware. Once ready, hit a key')
        print('and make it send as many RF data as possible.')
        print('Hit CTL-C to exit the scanner.')
        try:
            while True:
                for freq in self.__frequencies:
                    self.__phy.set_frequency(freq)
                    self.__current_freq = freq
                    sleep(self.__delay)
        except KeyboardInterrupt as kbd_err:
            # frequency sweep done, display statistics
            print('Frequency scan stopped by user, here are some statistics:')
            self.display_stats()

    def on_packet(self, packet):
        """We received a packet with a valid RSSI

        :param packet: Packet received
        :type packet: Packet
        """
        # Count this packet in our stats
        self.__stats[self.__current_freq] += 1

        # Display the packet
        pass

    def display_stats(self):
        """Show statistics in terminal
        """
        pass


class PhyInspector(CommandLineApp):

    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description='PHY inspection and analysis tool',
            interface=True,
            commands=False
        )
        self.add_argument(
            '--ble',
            action='store_true',
            dest='ble',
            default=False,
            help='Use BLE PHY (1Mbps, 250kHz deviation)'
        )

        self.add_argument(
            '-f',
            '--freq',
            type=int,
            dest='freq',
            help='Frequency (in Hz)'
        )

        self.add_argument(
            '-r',
            '--rssi',
            dest='rssi',
            type=int,
            default=-100,
            help='RSSI threshold. Only keep packets with RSSI above this value.'
        )

        ######################
        # Analysis modes
        ######################

        self.add_argument(
            '-sf',
            '--search-freq',
            dest='search_freq',
            metavar='FREQ_START:FREQ_END:STEP',
            help='''Search for frequency used to send a packet knowing the modulation scheme.
                    This mode sweeps a range of frequencies and listens for packet and will report
                    those with a strong RSSI specified with the --rssi command-line option.'''
        )

        self.add_argument(
            '-m',
            '--match',
            dest='pattern_match',
            help='''Search hex pattern in received packets, no matter what endianness or whitening used.
            Matching packets are displayed as well as guessed configuration. Patterns may contain wildcards
            noted as ??.
            '''
        )

        self.add_argument(
            '-s',
            '--similar',
            action='store_true',
            default=False,
            help='''Receive raw packets and find the longest common sequence of bytes.
            '''
        )


    def run(self):
        """Override App's run() method to handle scripting feature.
        """
        try:
            # Launch pre-run tasks
            self.pre_run()

            # We need to have an interface specified
            if self.interface is not None:
               # Process options (main)
               if self.args.ble:
                   # Configure hardware for BLE 1Mbps
                   self.analyze(ble_1mbps=True, rssi=self.args.rssi, freq=self.args.freq)
            else:
                self.error('You need to specify an interface with option --interface.')

        except KeyboardInterrupt as keybd:
            self.warning('phy-inspect stopped (CTL-C)')

        # Launch post-run tasks
        self.post_run()

    def analyze(self, freq=None, rssi=-100, ble_1mbps=False):
        phy = PhyCorrelator(self.interface, rssi=rssi)
        if ble_1mbps:
            print('rssi min: %d' % rssi)
            print('freq: %d' % freq)
            # Configure packet correlator
            phy.set_frequency(freq)
            phy.set_gfsk(deviation=250000)
            phy.set_packet_size(50)
            phy.set_sync_word(b'\xAA\xAA')
            phy.sniff_phy()

            # Start sniffing
            phy.start()

            input()
        else:
            self.warning('BLE 1Mbps is the only supported mode')



def phy_inspect_main():
    try:
        app = PhyInspector()
        app.run()
    except ApplicationError as err:
        err.show()



"""
This module provides a :class:`Sniffer` class that derives from :class:`ESB`
to allow *Enhanced ShockBurst* frame sniffing. Sniffed frames are wrapped in
a dedicated *Scapy* layer and therefore are represented as *Scapy* packets.

This :class:`Sniffer` class is used by the generic `wsniff` utility when
*Enhanced ShockBurst* protocol is selected, and the associated
:class:`whad.esb.sniffing.SnifferConfiguration` class holds all the sniffing
parameters that can be set.

By default, the *Enhanced ShockBurst* sniffer sniffs packet on all channels by
looping from channel 0 to 100 over and over and capturing frames that match
the expected format.
"""
from typing import Generator

from scapy.packet import Packet

from whad.device import Device
from whad.exceptions import WhadDeviceDisconnected, UnsupportedCapability
from whad.common.sniffing import EventsManager

from whad.hub.esb import PduReceived, RawPduReceived
from whad.hub.message import AbstractPacket

from .base import ESB
from ..sniffing import SnifferConfiguration


class Sniffer(ESB, EventsManager):
    """
    Enhanced ShockBurst Sniffer interface for compatible WHAD device.
    """

    def __init__(self, device: Device):
        ESB.__init__(self, device)
        EventsManager.__init__(self)

        self.__configuration = SnifferConfiguration()

        # Check if device can perform sniffing
        if not self.can_sniff():
            raise UnsupportedCapability("Sniff")

    def _enable_sniffing(self):
        """Configure the underlying ESB connector.


        This method parses the current configuration and sets the underlying
        ESB connector properties accordingly before starting sniffing packets.
        """
        if self.__configuration.scanning:
            channel = None
        else:
            channel = self.__configuration.channel

        ack = self.__configuration.acknowledgements
        address = self.__configuration.address

        super().start_sniff(channel=channel, show_acknowledgements=ack, address=address)
        self.start()

    @property
    def configuration(self) -> SnifferConfiguration:
        """Current configuration getter.
        """
        return self.__configuration

    @configuration.setter
    def configuration(self, new_configuration):
        """Configuration setter.

        :param new_configuration: New configuration to apply.
        :type new_configuration: :class:`whad.esb.sniffing.SnifferConfiguration`
        """
        self.stop()
        self.__configuration = new_configuration
        self._enable_sniffing()

    @property
    def channel(self) -> int:
        """Channel getter.
        """
        return self.__configuration.channel

    @channel.setter
    def channel(self, channel=11):
        """Channel setter.

        This method stops sniffing, update the current channel to the one
        specified into the configuration and then return to sniffing on this
        newly set channel.
        """
        self.stop()
        self.__configuration.channel = channel
        self._enable_sniffing()


    def sniff(self, channel: int = None, address: str = None, show_acknowledgements: bool = False,
              timeout: float = None) -> Generator[Packet, None, None]:
        """Update current sniffing configuration if required and sniff packets. This function
        returns after `timeout` seconds, if specified.

        :param channel: Channel to listen, None to iterate over all possible channels
        :type channel: int
        :param address: Device address to target
        :type address: str
        :param show_acknowledgements: Sniff packets acknowledgements if set to True (default: False)
        :type show_acknowledgements: bool
        :param timeout: Number of seconds after which sniffing will stop,
                        uninterrupted if set to None
        :type timeout: float
        """
        # Set channel, address and show ack if provided
        update_config_required = False
        if show_acknowledgements is not None:
            self.__configuration.acknowledgements = show_acknowledgements
            update_config_required = True
        if address is not None:
            self.__configuration.address = address
            update_config_required = True
        if channel is not None:
            self.__configuration.channel = channel

        # Update configuration if at least one parameter has been changed
        if update_config_required:
            self.stop()
            self._enable_sniffing()

        # Determine message type
        if self.support_raw_pdu():
            message_type = RawPduReceived
        else:
            message_type = PduReceived

        try:
            for message in super().sniff(messages=(message_type), timeout=timeout):
                if message is not None and issubclass(message, AbstractPacket):
                    packet = message.to_packet()
                    if packet is not None:
                        self.monitor_packet_rx(packet)
                        yield packet

        # Handle device disconnection
        except WhadDeviceDisconnected:
            pass



from whad.device import WhadDevice, WhadDeviceConnector
from whad.zigbee.connector.enddevice import EndDevice

from prompt_toolkit import print_formatted_text, HTML

from whad.common.monitors import WiresharkMonitor
from whad.exceptions import ExternalToolNotFound

from whad.cli.shell import InteractiveShell, category
from whad.dot15d4.exceptions import InvalidDot15d4AddressException
from whad.dot15d4.address import Dot15d4Address
from whad.zigbee.profile.nodes import CoordinatorNode, RouterNode, EndDeviceNode, \
    EndpointsDiscoveryException
from whad.zigbee.stack.apl.constants import ZIGBEE_PROFILE_IDENTIFIERS, \
    ZIGBEE_DEVICE_IDENTIFIERS, ZIGBEE_CLUSTER_IDENTIFIERS

from whad.zigbee.stack.apl.zcl import ZCLClientCluster
from whad.zigbee.profile.network import JoiningForbidden, NotAssociated, NotAuthorized
from .cache import ZigbeeNetworksCache
from .helpers import create_enddevice

INTRO='''
zigbee-enddevice, the WHAD Zigbee end device utility
'''

class ZigbeeEndDeviceShell(InteractiveShell):
    """Zigbee End Device interactive shell
    """

    def __init__(self, interface: WhadDevice = None, connector=None, network_panid=None):
        super().__init__(HTML('<b>zigbee-enddevice></b> '))

        # If interface is None, pick the first matching our needs
        self.__interface = interface
        self.__cache = ZigbeeNetworksCache()
        self.__wireshark = None

        # If connector is not provided
        if connector is None:
            # Reset target info and connector.
            self.__target_network = None
            self.__target_network_panid = None
            self.__connector: WhadDeviceConnector = EndDevice(self.__interface)
        else:
            # If connector provided, consider the network already connected
            self.__connector = connector
            self.__target_network = None
            self.__target_network_panid = network_panid

        self.intro = INTRO

        self.update_prompt()


    def update_prompt(self, force=False):
        """Update prompt to reflect current state
        """
        if not self.__target_network:
            self.set_prompt(HTML('<b>zigbee-enddevice></b> '), force)
        else:
            self.set_prompt(HTML('<b>zigbee-enddevice|<ansicyan>%s|%s</ansicyan>></b> ' % (
                    hex(self.__target_network['info'].pan_id),
                    str(Dot15d4Address(self.__target_network['info'].extended_pan_id))
                )
            ), force)


    def complete_wireshark(self):
        """Autocomplete wireshark command
        """
        completions = {}
        if self.__wireshark is not None:
            completions['off'] = {}
        else:
            completions['on'] = {}
        return completions


    @category("Monitoring")
    def do_wireshark(self, arg):
        """launch wireshark to monitor packets

        <ansicyan><b>wireshark</b> <i>["on" | "off"]</i></ansicyan>

        This command launches a wireshark that will display all the packets sent
        and received in the active connection.
        """
        if len(arg) >=1:
            enabled = arg[0].lower()=="on"
            if enabled:
                if self.__wireshark is None:
                    try:
                        self.__wireshark = WiresharkMonitor()
                        if self.__connector is not None:
                            self.__wireshark.attach(self.__connector)
                            self.__wireshark.start()
                    except ExternalToolNotFound:
                        self.error("Cannot launch Wireshark, please make sure it is installed.")
                else:
                    self.error("Wireshark is already launched, see wireshark off")
            else:
                # Detach monitor if any
                if self.__wireshark is not None:
                    self.__wireshark.detach()
                    self.__wireshark.close()
                    self.__wireshark = None
        else:
            self.error("Missing arguments, see help wireshark.")

    @category('Networks discovery')
    def do_scan(self, args):
        """scan surrounding networks and show a small summary

        <ansicyan><b>scan</b></ansicyan>

        Scan surrounding networks and report them in this console in real-time.

        The following information is provided:
         - <b>Channel:</b> represents the channel where the network is deployed
         - <b>Pan ID:</b> short identifier of the ZigBee network
         - <b>Ext. Pan ID:</b> long (extended) identifier of the ZigBee network
         - <b>Joining:</b> indicates if joining the network is allowed or not

        You can stop a scan by hitting <b>CTL-c</b> at any time, the discovered networks are kept in
        memory and would be available in autocompletion.
        """

        # Start scanning
        print_formatted_text(HTML('<ansigreen>Channel   PAN ID   Ext. PAN ID             Joining</ansigreen>'))
        self.__connector.start()
        try:
            for network in self.__connector.discover_networks():
                # Show network
                print(
                        network.channel," "*6,
                        hex(network.pan_id)," ",
                        Dot15d4Address(network.extended_pan_id),
                        "permitted" if network.is_joining_permitted() else "forbidden"
                )

                # Add network to cache
                self.__cache.add(network)
        except KeyboardInterrupt as keybd_int:
            print('\rScan terminated by user')

        if self.__wireshark is not None:
            self.__wireshark.detach()

        self.__connector.stop()


    @category('Networks discovery')
    def do_networks(self, arg):
        """list discovered networks

        <ansicyan><b>networks</b></ansicyan>

        List every discovered networks so far, through the <ansicyan>scan</ansicyan> command.
        This command displays the content of the console networks cache.
        """
        print_formatted_text(HTML('<ansigreen>Channel   PAN ID   Ext. PAN ID             Joining</ansigreen>'))
        for network in self.__cache.iterate():
            # Show network
            print(
                    network['info'].channel," "*6,
                    hex(network['info'].pan_id)," ",
                    Dot15d4Address(network['info'].extended_pan_id),
                    "permitted" if network['info'].is_joining_permitted() else "forbidden"
            )



    def get_cache_targets(self):
        # Keep track of Extended PAN ID
        targets = [str(Dot15d4Address(network['info'].extended_pan_id)) for network in self.__cache.iterate()]
        targets.extend(['%s' % hex(network['info'].pan_id) for network in self.__cache.iterate()])
        return targets

    def complete_join(self):
        """Autocomplete the 'join' command, providing extended PAN ID of surrounding networks.
        """
        # Keep track of Extended PAN ID
        targets = self.get_cache_targets()
        completions = self.autocomplete_env()
        for address in targets:
            completions[address] = None
        return completions

    @category('Network interaction')
    def do_join(self, args):
        """join a network

        <ansicyan><b>join</b> <i>[ Extended PAN ID or PAN ID ] (force)</i></ansicyan>

        Initiate a ZigBee join to a specific network by its extended PAN ID or
        PAN ID. If multiple networks have the same PAN ID, the first one will be
        picked for join.

        Example:

            <i>    zigbee-enddevice> join 0x1234</i>

        Joining a network that des not permit association is possible by using the
        optional <i>force</i> parameter.

        Example:

            <i>    zigbee-enddevice> join 0x1234 force</i>
        """

        if len(args) < 1:
            self.error('<u>join</u> requires at least one parameter (extended PAN ID or PAN ID).\ntype \'help join\' for more details.')
            return

        #try:
        target = None
        force = len(args) >= 2 and args[1].lower() == "force"

        try:
            target = self.__cache[args[0]]
            target_pan_id = Dot15d4Address(target['info'].extended_pan_id)

        except (IndexError, KeyError):
            # If target not in cache, we are expecting an extended PAN ID or a PAN ID
            try:
                target_pan_id = Dot15d4Address(args[0])
            except InvalidDot15d4AddressException:
                self.error('You must provide a valid extended PAN ID or PAN ID.')
                return

        # Look in cache by PAN ID
        if target is None:
            for network in self.__cache.iterate():
                if target_pan_id == network['info'].pan_id:
                    target = network
                    break

        if target is None:
            # Discover surrounding networks if needed
            for network in self.__connector.discover_networks():
                # Add network to cache
                if target_pan_id == network.extended_pan_id or target_pan_id == network.pan_id:
                    self.__cache.add(network)
                    target = self.__cache[target_pan_id]
                    break

            if target is None:
                self.error('No matching network found.')
                return

        try:
            # Attach our wireshark monitor, if any
            if self.__wireshark is not None:
                self.__wireshark.attach(self.__connector)

            print("Joining target network (PAN ID = %s / Ext. PAN ID = %s) on channel %d ..." % (
                hex(target['info'].pan_id),
                str(Dot15d4Address(target['info'].extended_pan_id)),
                target['info'].channel
                )
            )

            # Perform join
            result = target['info'].join(force=force)

            if result:
                print("Successfully joined to target network (PAN ID = %s / Ext. PAN ID = %s)." % (
                        hex(target['info'].pan_id),
                        str(Dot15d4Address(target['info'].extended_pan_id))
                    )
                )
            else:
                self.error("Failed to join target network (PAN ID = %s / Ext. PAN ID = %s)." % (
                        hex(target['info'].pan_id),
                        str(Dot15d4Address(target['info'].extended_pan_id))
                    )
                )

            if target['info'].network_key is not None:
                print_formatted_text(HTML('<ansicyan><b>Network key:</b></ansicyan> %s' % target['info'].network_key.hex()))

                self.__target_network = target
                # Update prompt
                self.update_prompt()

        except JoiningForbidden:
            self.error('Joining forbidden.')


    @category('Network interaction')
    def do_nodes(self, args):
        """discover nodes present within the associated network.

        <ansicyan><b>nodes</b></ansicyan>

        This command performs a nodes discovery, collecting all this information and
        keeping it in a dedicated <b>cache</b>.

        <aaa fg="orange">Sometimes this discovery process may cause an error and produces
        incomplete information, in this case try again and cross fingers.</aaa>
        """
        if self.__target_network is not None:
            if not self.__target_network['discovered']:
                print("Discovering surrounding nodes.")
                nodes = self.__target_network['info'].discover()
                self.__target_network['discovered'] = True
                for node in nodes:
                    if isinstance(node, CoordinatorNode):
                        print("New Coordinator discovered (addr. = %s, ext. addr. = %s)" % (
                                str(Dot15d4Address(node.address)),
                                str(Dot15d4Address(node.extended_address))
                            )
                        )
                    elif isinstance(node, RouterNode):
                        print("New Router discovered (addr. = %s, ext. addr. = %s)" % (
                                str(Dot15d4Address(node.address)),
                                str(Dot15d4Address(node.extended_address))
                            )
                        )
                    else:
                        print("New End Device discovered (addr. = %s, ext. addr. = %s)" % (
                                str(Dot15d4Address(node.address)),
                                str(Dot15d4Address(node.extended_address))
                            )
                        )

            print_formatted_text(HTML('<ansigreen>Addr.     Ext. addr.                  Type</ansigreen>'))

            for node in self.__target_network['info'].nodes:
                if isinstance(node, CoordinatorNode):
                    print("%s    %s     coordinator" % (
                            str(Dot15d4Address(node.address)),
                            str(Dot15d4Address(node.extended_address))
                        )
                    )
                elif isinstance(node, RouterNode):
                    print("%s    %s     router" % (
                            str(Dot15d4Address(node.address)),
                            str(Dot15d4Address(node.extended_address))
                        )
                    )
                else:
                    print("%s    %s     end device" % (
                            str(Dot15d4Address(node.address)),
                            str(Dot15d4Address(node.extended_address))
                        )
                    )
        else:
            self.error("You must join a network to perform a discovery.")
            return



    def get_cache_nodes(self):
        # Keep track of Nodes
        if self.__target_network is None:
            return []
        nodes = []
        if self.__target_network['discovered']:
            nodes = ["0x{:04x}".format(node.address) for node in self.__target_network['info'].nodes]
            nodes.extend(['%s' % str(Dot15d4Address(node.extended_address)) for node in self.__target_network['info'].nodes])
        return nodes

    def complete_endpoints(self):
        """Autocomplete the 'endpoints' command, providing address or extended address of neighbors nodes.
        """
        # Keep track of Extended PAN ID
        nodes = self.get_cache_nodes()
        completions = self.autocomplete_env()
        for address in nodes:
            completions[address] = None
        return completions

    @category('Node interaction')
    def do_endpoints(self, args):
        """discover endpoints related to a given node.

        <ansicyan><b>endpoints</b> <i>[addr. | ext. addr]</i></ansicyan>

        This command performs a endpoints discovery, collecting all this information and
        keeping it in a dedicated <b>cache</b>.

        <aaa fg="orange">Sometimes this discovery process may cause an error and produces
        incomplete information, in this case try again and cross fingers.</aaa>
        """
        if self.__target_network is not None:
            if len(args) < 1:
                self.error('No target node provided. ')
                return

            try:
                target_node = Dot15d4Address(args[0])
            except InvalidDot15d4AddressException:
                self.error("Invalid 802.15.4 address.")
                return

            if not self.__target_network['discovered']:
                print("Discovering surrounding nodes.")
                nodes = self.__target_network['info'].discover()
                self.__target_network['discovered'] = True

            selected_node = None
            for node in self.__target_network['info'].nodes:
                if (
                    target_node == Dot15d4Address(node.address) or
                    target_node == Dot15d4Address(node.extended_address)
                ):
                    selected_node = node
                    break

            if selected_node is None:
                self.error("No matching node found.")
                return
            try:
                for endpoint in selected_node.endpoints:
                    try:
                        profile_name = ZIGBEE_PROFILE_IDENTIFIERS[endpoint.profile_id]
                    except (IndexError, KeyError):
                        profile_name = "Unknown"

                    try:
                        device_name = ZIGBEE_DEVICE_IDENTIFIERS[endpoint.device_id]
                    except (IndexError, KeyError):
                        device_name = "Unknown device"

                    print_formatted_text(HTML('<ansigreen><b>Endpoint #%d: %s profile (%s) -> %s device (%s) </b></ansigreen>' %
                            (
                                endpoint.number,
                                profile_name,
                                hex(endpoint.profile_id),
                                device_name,
                                hex(endpoint.device_id)
                            )
                        )
                    )
                    print_formatted_text(HTML('  | <ansicyan><b>Input clusters:</b></ansicyan>'))
                    for cluster_id in endpoint.input_clusters:
                        try:
                            category, cluster_name = ZIGBEE_CLUSTER_IDENTIFIERS[cluster_id]
                        except (IndexError, KeyError):
                            category, cluster_name = ("Unknown category", "Unknown cluster")
                        print_formatted_text(HTML('    | <b>%s - %s (%s)</b>' % (category, cluster_name, hex(cluster_id))))

                    print_formatted_text(HTML('  | <ansicyan><b>Output clusters:</b></ansicyan>'))
                    for cluster_id in endpoint.output_clusters:
                        try:
                            category, cluster_name = ZIGBEE_CLUSTER_IDENTIFIERS[cluster_id]
                        except (IndexError, KeyError):
                            category, cluster_name = ("Unknown category", "Unknown cluster")
                        print_formatted_text(HTML('    | <b>%s - %s (%s)</b>' % (category, cluster_name, hex(cluster_id))))
                print()

            except EndpointsDiscoveryException:
                self.error("An error occurred during endpoints discovery.")
                return
        else:
            self.error("You must join a network to perform a discovery.")
            return

    def get_cache_endpoints(self, address):
        input_clusters = {}
        for node in self.__target_network['info'].nodes:
            if address == "0x{:04x}".format(node.address) or address == str(Dot15d4Address(node.extended_address)).lower():
                endpoints = {}
                try:
                    for endpoint in node.endpoints:
                        input_clusters = {}
                        if endpoint is not None:
                            for input_cluster in endpoint.input_clusters:
                                commands = {}
                                for candidate_cluster_class in ZCLClientCluster.child_clusters():
                                    candidate = candidate_cluster_class()
                                    if candidate.cluster_id == input_cluster:
                                        for command_id, command in candidate.CLUSTER_SPECIFIC_COMMANDS.items():
                                            commands[str(command['name'].replace(' ', '_').lower())] = None
                                        for command_id, command in candidate.PROFILE_WIDE_COMMANDS.items():
                                            commands[str(command['name'].replace(' ', '_').lower())] = None
                                        break
                                input_clusters[str(input_cluster)] = commands
                        endpoints[str(endpoint.number)] = input_clusters
                except EndpointsDiscoveryException:
                    pass
        return endpoints

    def complete_cluster_cmd(self):
        """Autocomplete the 'cluster_cmd' command, providing address or extended address of neighbors nodes.
        """
        # Keep track of Extended PAN ID
        nodes = self.get_cache_nodes()
        completions = self.autocomplete_env()
        for address in nodes:
            input_clusters = self.get_cache_endpoints(address)
            '''
            for node in self.__target_network['info'].nodes:
                if address == "0x{:04x}".format(node.address) or address == str(Dot15d4Address(node.extended_address)).lower():
                    for endpoint in node.endpoints:
                        for input_cluster in endpoint.input_clusters:
                            input_clusters[str(input_cluster)] = None
                            try:
                                category, cluster_name = ZIGBEE_CLUSTER_IDENTIFIERS[input_cluster]
                                cluster_name = cluster_name.replace(" ", "_").replace("/","").lower()
                                input_clusters[cluster_name] = None

                            except IndexError:
                                pass
            '''
            completions[address] = input_clusters
        return completions

    @category('Node interaction')
    def do_cluster_cmd(self, args):
        """Run a cluster command on a given node.

        <ansicyan><b>cluster_cmd</b> <i>[addr. | ext. addr] [cluster id | cluster name] [cmd id | cmd name] [cmd parameters]</i></ansicyan>

        This command performs a cluster command targeted on a given name.
        """
        if self.__target_network is not None:
            if len(args) < 1:
                self.error('No target node provided. ')
                return

            if len(args) < 2:
                self.error('No target endpoint provided. ')
                return

            if len(args) < 3:
                self.error('No target cluster provided. ')
                return

            if len(args) < 4:
                self.error('No target cluster command provided. ')
                return

            try:
                target_node = Dot15d4Address(args[0])
            except InvalidDot15d4AddressException:
                self.error("Invalid 802.15.4 address.")
                return

            try:
                target_endpoint = int(args[1])
            except:
                self.error("Invalid endpoint.")
                return

            try:
                target_cluster = int(args[2])
            except:
                self.error("Invalid cluster.")
                return


            try:
                target_cmd = args[3]
                if len(args) >= 4:
                    target_cmd_parameters = args[4:]
                else:
                    target_cmd_parameters = None
            except:
                self.error("Invalid command.")
                return


            if not self.__target_network['discovered']:
                print("Discovering surrounding nodes.")
                nodes = self.__target_network['info'].discover()
                self.__target_network['discovered'] = True

            selected_node = None
            for node in self.__target_network['info'].nodes:
                if (
                    target_node == Dot15d4Address(node.address) or
                    target_node == Dot15d4Address(node.extended_address)
                ):
                    selected_node = node
                    break

            if selected_node is None:
                self.error("No matching node found.")
                return
            try:
                selected_endpoint = None
                for endpoint in selected_node.endpoints:
                    if endpoint.number == target_endpoint:
                        selected_endpoint = endpoint
                        break

                if selected_endpoint is None:
                    self.error("No matching endpoint found.")
                    return

                cluster = None
                if target_cluster in selected_endpoint.input_clusters:
                    cluster = selected_endpoint.attach_to_input_cluster(target_cluster)

                if cluster is None:
                    self.error("No matching cluster or cluster not implemented.")
                    return

                selected_command_name = None
                for command_id, command in cluster.PROFILE_WIDE_COMMANDS.items():
                    if target_cmd == command['name'].replace(" ","_").lower():
                        selected_command_name = command['generate_callback'].__name__

                for command_id, command in cluster.CLUSTER_SPECIFIC_COMMANDS.items():
                    if target_cmd == command['name'].replace(" ","_").lower():
                        selected_command_name = command['generate_callback'].__name__

                if selected_command_name is None:
                    self.error("No matching command found.")
                    return

                getattr(cluster, selected_command_name)()

            except EndpointsDiscoveryException:
                self.error("An error occurred during cluster command.")
                return
        else:
            self.error("You must join a network to perform a cluster command.")
            return


    def do_quit(self, args):
        """close wzb-enddevice
        """
        if self.__connector is not None:
            self.__connector.stop()
        if self.__interface is not None:
            self.__interface.close()
        return True

    def do_exit(self, arg):
        """alias for <ansicyan>quit</ansicyan>
        """
        return self.do_quit(arg)



from whad.zigbee.stack.nwk.nodes import ZigbeeEndDevice, ZigbeeCoordinator, ZigbeeRouter
from whad.zigbee.stack.nwk.constants import ZigbeeDeviceType, ZigbeeRelationship

class NWKNeighborTable:
    """
    Structure representing a NWK neighbor table, to store the characteristics of surrounding devices.
    """
    def __init__(self):
        self.table = {}

    def update(self, address, device_type, extended_pan_id, **kwargs):
        kwargs["device_type"] = device_type
        kwargs["extended_pan_id"] = extended_pan_id

        if (address, extended_pan_id) in self.table:
            device = self.table[(address, extended_pan_id)]
            for name, value in kwargs.items():
                if hasattr(device, name):
                    setattr(device, name, value)
            return True
        else:
            if "device_type" not in kwargs:
                return False
            if kwargs["device_type"] == ZigbeeDeviceType.END_DEVICE:
                del kwargs["device_type"]
                self.table[(address, extended_pan_id)] = ZigbeeEndDevice(address, **kwargs)
            elif kwargs["device_type"] == ZigbeeDeviceType.COORDINATOR:
                del kwargs["device_type"]
                self.table[(address, extended_pan_id)] = ZigbeeCoordinator(address, **kwargs)
            elif kwargs["device_type"] == ZigbeeDeviceType.ROUTER:
                del kwargs["device_type"]
                self.table[(address, extended_pan_id)] = ZigbeeRouter(address, **kwargs)
            else:
                return False
            return True

    def select_by_extended_address(self, extended_address):
        for (address, extended_pan_id), node in self.table.items():
            if node.extended_address == extended_address:
                return node
        return None

    def select_routers_by_pan_id(self, pan_id):
        routers = []
        for (address, extended_pan_id), device in self.table.items():
            if device.device_type == ZigbeeDeviceType.ROUTER and device.pan_id == pan_id:
                routers.append(device)
        return routers

    def select_end_devices_by_pan_id(self, pan_id):
        end_devices = []
        for (address, extended_pan_id), device in self.table.items():
            if device.device_type == ZigbeeDeviceType.END_DEVICE and device.pan_id == pan_id:
                end_devices.append(device)
        return end_devices

    def select_suitable_parent(self, extended_pan_id, nwk_update_id, no_permit_check=False):
        selected_devices = []
        for (address, extended_pan_id), device in self.table.items():
            if (
                device.extended_pan_id == extended_pan_id and #the device belongs to the right network
                (device.permit_joining or no_permit_check) and  # the device allows joining
                device.outgoing_cost <= 3 and # the total cost is under 3
                device.potential_parent and # it is a potential parent
                device.update_id >= nwk_update_id
            ):
                selected_devices.append(device)
        return selected_devices

    def get_parent(self):
        for device in self.table.values():
            if device.relationship == ZigbeeRelationship.IS_PARENT:
                return device
        return None

    def delete(self, address, extended_pan_id):
        del self.table[(address, extended_pan_id)]

    def show(self):
        for _, device in self.table.items():
            print(device)



"""
This module provides some constants used by WHAD to communicate with the Yard Stick One.
"""
from enum import IntEnum
from dataclasses import dataclass


class YardInternalStates(IntEnum):
    """Internal states.
    """
    YARD_MODE_IDLE = 0
    YARD_MODE_RX = 1
    YARD_MODE_TX = 2

@dataclass
class RegisterMask:
    """Register mask class.

    This class stores the offset to a register and its
    associated mask.
    """
    mask : int = 0
    offset : int = 0

class YardUSBProperties:
    """USB timeouts.
    """
    USB_RX_WAIT = 1000
    USB_TX_WAIT = 10000

class YardStickOneId(IntEnum):
    """Yard USB vendor and product IDs.
    """
    YARD_ID_VENDOR    = 0x1d50
    YARD_ID_PRODUCT   = 0x605b

class YardStickOneEndPoints(IntEnum):
    """USB endpoints.
    """
    SETUP_ENDPOINT      = 0x00
    IN_ENDPOINT         = 0x85
    OUT_ENDPOINT        = 0x05

class YardApplications(IntEnum):
    """Applications.
    """
    GENERIC                 = 0x01
    DEBUG                   = 0xfe
    SYSTEM                  = 0xff
    NIC                     = 0x42
    SPECAN                  = 0x43

class YardSystemCommands(IntEnum):
    """System commands op-codes.
    """
    PEEK                    = 0x80
    POKE                    = 0x81
    PING                    = 0x82
    STATUS                  = 0x83
    POKE_REG                = 0x84
    GET_CLOCK               = 0x85
    BUILDTYPE               = 0x86
    BOOTLOADER              = 0x87
    RFMODE                  = 0x88
    COMPILER                = 0x89
    PARTNUM                 = 0x8e
    RESET                   = 0x8f
    CLEAR_CODES             = 0x90
    DEVICE_SERIAL_NUMBER    = 0x91
    LED_MODE                = 0x93

class YardNICCommands(IntEnum):
    """NICC commands.
    """
    RECV                = 0x1
    XMIT                = 0x2
    SET_ID              = 0x3
    SET_RECV_LARGE      = 0x5
    SET_AES_MODE        = 0x6
    GET_AES_MODE        = 0x7
    SET_AES_IV          = 0x8
    SET_AES_KEY         = 0x9
    SET_AMP_MODE        = 0xa
    GET_AMP_MODE        = 0xb
    LONG_XMIT           = 0xc
    LONG_XMIT_MORE      = 0xd

class YardRFStates:
    """RF states
    """
    SFSTXON                 = 0x00
    SCAL                    = 0x01
    SRX                     = 0x02
    STX                     = 0x03
    SIDLE                   = 0x04
    SNOP                    = 0x05

class YardMARCStates(IntEnum):
    """Main Radio Control State.
    """
    MARC_STATE_SLEEP               = 0x00
    MARC_STATE_IDLE                = 0x01
    MARC_STATE_VCOON_MC            = 0x03
    MARC_STATE_BWBOOST             = 0x09
    MARC_STATE_ENDCAL              = 0x0C
    MARC_STATE_FSTXON              = 0x12
    MARC_STATE_FS_LOCK             = 0x0A
    MARC_STATE_IFADCON             = 0x0B
    MARC_STATE_MANCAL              = 0x05
    MARC_STATE_REGON               = 0x07
    MARC_STATE_REGON_MC            = 0x04
    MARC_STATE_RX                  = 0x0D
    MARC_STATE_RXTX_SWITCH         = 0x15
    MARC_STATE_RX_END              = 0x0E
    MARC_STATE_RX_OVERFLOW         = 0x11
    MARC_STATE_RX_RST              = 0x0F
    MARC_STATE_STARTCAL            = 0x08
    MARC_STATE_TX                  = 0x13
    MARC_STATE_TXRX_SWITCH         = 0x10
    MARC_STATE_TX_END              = 0x14
    MARC_STATE_TX_UNDERFLOW        = 0x16
    MARC_STATE_VCOON               = 0x06

class YardRadioStructure:
    """Radio configuration structure.
    """
    FIELDS = [
        "SYNC1",
        "SYNC0",
        "PKTLEN",
        "PKTCTRL1",
        "PKTCTRL0",
        "ADDR",
        "CHANNR",
        "FSCTRL1",
        "FSCTRL0",
        "FREQ2",
        "FREQ1",
        "FREQ0",
        "MDMCFG4",
        "MDMCFG3",
        "MDMCFG2",
        "MDMCFG1",
        "MDMCFG0",
        "DEVIATN",
        "MCSM2",
        "MCSM1",
        "MCSM0",
        "FOCCFG",
        "BSCFG",
        "AGCCTRL2",
        "AGCCTRL1",
        "AGCCTRL0",
        "FREND1",
        "FREND0",
        "FSCAL3",
        "FSCAL2",
        "FSCAL1",
        "FSCAL0",
        "Z0",
        "Z1",
        "Z2",
        "TEST2",
        "TEST1",
        "TEST0",
        "Z3",
        "PA_TABLE7",
        "PA_TABLE6",
        "PA_TABLE5",
        "PA_TABLE4",
        "PA_TABLE3",
        "PA_TABLE2",
        "PA_TABLE1",
        "PA_TABLE0",
        "IOCFG2",
        "IOCFG1",
        "IOCFG0",
        "Z4",
        "Z5",
        "Z6",
        "Z7",
        "PARTNUM",
        "CHIPID",
        "FREQEST",
        "LQI",
        "RSSI",
        "MARCSTATE",
        "PKSTATUS",
        "VCO_VC_DAC"
    ]

    BASE_ADDRESS = 0xDF00
    MEMORY_SIZE = 0x3E

    def __init__(self, poke, peek):
        self._poke = poke
        self._peek = peek

        self._memory = self._peek(YardRadioStructure.BASE_ADDRESS,YardRadioStructure.MEMORY_SIZE)

    def update(self):
        """Read current configuration from memory.
        """
        self._memory = self._peek(YardRadioStructure.BASE_ADDRESS,YardRadioStructure.MEMORY_SIZE)

    def get(self, name: str) -> int:
        """Get a specific configuration field from current configuration.

        :param name: Configuration field
        :type name: str
        :return: Field value
        :rtype: bytes
        """
        offset = YardRadioStructure.FIELDS.index(name)
        return self._memory[offset]

    def set(self, name: str, data: int) -> bool:
        """Set a specific field value in current configuration.

        :param name: Field name
        :type name: str
        :param data: Field value
        :type data: int
        :return: `True` if field has successfully been updated, `False` otherwise.
        :rtype: bool
        """
        offset = YardRadioStructure.FIELDS.index(name)
        self._poke(YardRadioStructure.BASE_ADDRESS + offset, bytes([data]))
        if self._peek(YardRadioStructure.BASE_ADDRESS + offset, 1) == bytes([data]):
            self._memory = self._memory[:offset] + bytes([data]) + self._memory[offset+1:]
            return True
        return False

    def __repr__(self) -> str:
        """Get the textual representation of the Yard configuration structure.

        :return: Configuration structure reprensentation
        :rtype: str
        """
        return "\n".join([f"{name} -> {self.get(name):02x}" for name in YardRadioStructure.FIELDS])


class YardMemoryRegisters(IntEnum):
    """Yard registers
    """
    RFST = 0xDFE1

class YardCCA(IntEnum):
    """Yard clear channel assessment (CCA)
    """
    NO_CCA                      = 0
    CCA_RSSI_THRESHOLD          = 1
    CCA_PACKET                  = 2
    CCA_PACKET_RSSI_THRESHOLD   = 3

class YardFrequencyTransitionPoints:
    """Frequency band and VCO transition points.
    """

    # band transition points in Hz
    FREQ_EDGE_400 = 369000000
    FREQ_EDGE_900 = 615000000

    # VCO transition points in Hz
    FREQ_MID_300  = 318000000
    FREQ_MID_400  = 424000000
    FREQ_MID_900  = 848000000

class YardVCOType:
    """Voltage Control Oscillator types.
    """
    LOW_VCO = 0x0A
    HIGH_VCO = 0x2A

class YardModulations:
    """Supported modulations.
    """
    MODULATION_2FSK                        = 0b000
    MODULATION_GFSK                        = 0b001
    MODULATION_ASK                         = 0b011
    MODULATION_4FSK                        = 0b100
    MODULATION_MSK                         = 0b111

class YardEncodings:
    """Supported encodings.
    """
    NON_RETURN_TO_ZERO  = 0
    MANCHESTER          = 1

# Yard Registers Masks
class YardRegistersMasks:
    """Registers masks.
    """

    class MDMCFG1:
        """Modem Config register 1
        """
        CHANSPC_E = RegisterMask(mask=0b11, offset=0)
        NUM_PREAMBLE = RegisterMask(mask=0b111, offset=4)
        FEC_EN = RegisterMask(mask=0b1, offset=7)


    class MDMCFG2:
        """Modem config register 2
        """
        SYNC_MODE = RegisterMask(mask=0b111, offset=0)
        MANCHESTER_EN = RegisterMask(mask=0b1, offset=3)
        MOD_FORMAT = RegisterMask(mask=0b111, offset=4)
        DEM_DCFILT_OFF = RegisterMask(mask=0b1, offset=7)


    class MDMCFG3:
        """Modem config register 3
        """
        DRATE_M = RegisterMask(mask=0b11111111, offset=0)

    class MDMCFG4:
        """Modem config register 4
        """
        DRATE_E = RegisterMask(mask=0b1111, offset=0)
        CHANBW_M = RegisterMask(mask=0b11, offset=4)
        CHANBW_E = RegisterMask(mask=0b11, offset=6)

    class PKTCTRL0:
        """Packet control register 0
        """
        LENGTH_CONFIG = RegisterMask(mask=0b11, offset=0)
        CRC_EN = RegisterMask(mask=0b1, offset=2)
        PKT_FORMAT = RegisterMask(mask=0b11, offset=4)
        WHITE_DATA = RegisterMask(mask=0b1, offset=6)


    class PKTCTRL1:
        """Packet control register 1
        """
        ADR_CHK = RegisterMask(mask=0b11, offset=0)
        APPEND_STATUS = RegisterMask(mask=0b1, offset=2)
        PQT = RegisterMask(mask=0b111, offset=5)

    class MCSM1:
        """Mode control registers.
        """
        TXOFF_MODE = RegisterMask(mask=0b11, offset=0)
        RXOFF_MODE = RegisterMask(mask=0b11, offset=2)
        CCA_MODE = RegisterMask(mask=0b11, offset=4)


    class MCSM2:
        """RX time registers.
        """
        RX_TIME = RegisterMask(mask=0b111, offset=0)
        RX_TIME_QUAL = RegisterMask(mask=0b1, offset=3)
        RX_TIME_RSSI = RegisterMask(mask=0b1, offset=4)

    class AGCCTRL1:
        """Adaptative Gain Control control register 1
        """
        CARRIER_SENSE_ABS_THR = RegisterMask(mask=0b1111, offset=0)
        CARRIER_SENSE_REL_THR = RegisterMask(mask=0b11, offset=4)
        AGC_LNA_PRIORITY = RegisterMask(mask=0b1, offset=6)


    class AGCCTRL2:
        """Adaptative Gain Control control register 2
        """
        MAGN_TARGET  = RegisterMask(mask=0b111, offset=0)
        MAX_LNA_GAIN  = RegisterMask(mask=0b111, offset=3)
        MAX_DVGA_GAIN  = RegisterMask(mask=0b11, offset=6)

    class DEVIATN:
        """Deviation control register
        """
        DEVIATION_M = RegisterMask(mask=0b111, offset=0)
        DEVIATION_E = RegisterMask(mask=0b111, offset=4)

    class FSCTRL0:
        """Frequency offset control register 0
        """
        FREQ_OFFSET = RegisterMask(mask=0b11111111, offset=0)

    class FSCTRL1:
        """Frequency offset control register 1
        """
        FREQ_IF = RegisterMask(mask=0b11111, offset=0)

    class FREND0:
        """Power amplifier control registers.
        """
        PA_POWER = RegisterMask(mask=0b111, offset=0)
        LODIV_BUF_CURRENT_TX = RegisterMask(mask=0b11, offset=4)

    class FREND1:
        """Low-noise amplifier control registers.
        """
        MIX_CURRENT = RegisterMask(mask=0b11, offset=0)
        LODIV_BUF_CURRENT_RX = RegisterMask(mask=0b11, offset=2)
        LNA2MIX_CURRENT = RegisterMask(mask=0b11, offset=4)
        LNA_CURRENT = RegisterMask(mask=0b11, offset=6)


# Number of preamble bytes lookup table
NUM_PREAMBLE_LOOKUP_TABLE = [
    2,
    3,
    4,
    6,
    8,
    12,
    16,
    24
]

# Possible bandwidth list
POSSIBLE_CHANNEL_BANDWIDTHS = [
    53e3,
    63e3,
    75e3,
    93e3,
    107e3,
    125e3,
    150e3,
    188e3,
    214e3,
    250e3,
    300e3,
    375e3,
    428e3,
    500e3,
    600e3,
    750e3,
]



"""WHAD Protocol Discovery speed message abstraction layer.
"""

from whad.hub.message import pb_bind, PbFieldInt, PbMessageWrapper
from whad.hub.discovery import Discovery

@pb_bind(Discovery, 'set_speed', 1)
class SetSpeed(PbMessageWrapper):
    """Device set speed message class
    """
    speed = PbFieldInt('discovery.set_speed.speed')


from enum import IntEnum


class PairingEntryStatus(IntEnum):
    """
    List of possible status for a pairing entry.
    """
    PROVISIONAL = 1
    ACTIVE = 2

class PairingEntry:
    """
    Defines a pairing table entry.
    """
    def __init__(self, source_network_address, channel, destination_ieee_address, destination_pan_id, destination_network_address, capabilities, frame_counter, link_key):
        self.source_network_address = source_network_address
        self.channel = channel
        self.destination_ieee_address= destination_ieee_address
        self.destination_network_address = destination_network_address
        self.destination_pan_id = destination_pan_id
        self.capabilities = capabilities
        self.frame_counter = frame_counter
        self.link_key = link_key
        self.status = PairingEntryStatus.PROVISIONAL


    def is_active(self):
        """
        Returns a boolean indicating if entry is active.
        """
        return self.status == PairingEntryStatus.ACTIVE
        
    def mark_as_active(self):
        """
        Mark this entry as active.
        """
        self.status = PairingEntryStatus.ACTIVE



class Dot15d4TimeoutException(Exception):
    """
    Default exception raised when a wait for packet method reachs its timeout.
    """
    def __init__(self):
        super().__init__()

class InvalidDot15d4AddressException(Exception):
    """
    Default exception raised when an invalid 802.15.4 address or Pan ID is provided.
    """
    def __init__(self):
        super().__init__()
        



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: whad/protocol/device.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1awhad/protocol/device.proto\x12\tdiscovery\"\x12\n\x10\x44\x65viceResetQuery\"\x11\n\x0f\x44\x65viceReadyResp\"\"\n\x11SetTransportSpeed\x12\r\n\x05speed\x18\x01 \x01(\r\"\xe0\x01\n\x0e\x44\x65viceInfoResp\x12\x0c\n\x04type\x18\x01 \x01(\r\x12\r\n\x05\x64\x65vid\x18\x02 \x01(\x0c\x12\x15\n\rproto_min_ver\x18\x03 \x01(\r\x12\x11\n\tmax_speed\x18\x04 \x01(\r\x12\x11\n\tfw_author\x18\x05 \x01(\x0c\x12\x0e\n\x06\x66w_url\x18\x06 \x01(\x0c\x12\x18\n\x10\x66w_version_major\x18\x07 \x01(\r\x12\x18\n\x10\x66w_version_minor\x18\x08 \x01(\r\x12\x16\n\x0e\x66w_version_rev\x18\t \x01(\r\x12\x18\n\x0c\x63\x61pabilities\x18\n \x03(\rB\x02\x10\x01\"B\n\x14\x44\x65viceDomainInfoResp\x12\x0e\n\x06\x64omain\x18\x01 \x01(\r\x12\x1a\n\x12supported_commands\x18\x02 \x01(\x04\"$\n\x0f\x44\x65viceInfoQuery\x12\x11\n\tproto_ver\x18\x01 \x01(\r\"\'\n\x15\x44\x65viceDomainInfoQuery\x12\x0e\n\x06\x64omain\x18\x01 \x01(\r\"\xfd\x02\n\x07Message\x12\x32\n\x0breset_query\x18\x01 \x01(\x0b\x32\x1b.discovery.DeviceResetQueryH\x00\x12\x30\n\nready_resp\x18\x02 \x01(\x0b\x32\x1a.discovery.DeviceReadyRespH\x00\x12\x30\n\ninfo_query\x18\x03 \x01(\x0b\x32\x1a.discovery.DeviceInfoQueryH\x00\x12.\n\tinfo_resp\x18\x04 \x01(\x0b\x32\x19.discovery.DeviceInfoRespH\x00\x12\x38\n\x0c\x64omain_query\x18\x05 \x01(\x0b\x32 .discovery.DeviceDomainInfoQueryH\x00\x12\x36\n\x0b\x64omain_resp\x18\x06 \x01(\x0b\x32\x1f.discovery.DeviceDomainInfoRespH\x00\x12\x31\n\tset_speed\x18\x07 \x01(\x0b\x32\x1c.discovery.SetTransportSpeedH\x00\x42\x05\n\x03msg*\xc6\x01\n\x06\x44omain\x12\x0f\n\x0b_DomainNone\x10\x00\x12\n\n\x03Phy\x10\x80\x80\x80\x08\x12\x10\n\tBtClassic\x10\x80\x80\x80\x10\x12\x0b\n\x04\x42tLE\x10\x80\x80\x80\x18\x12\x0e\n\x07\x44ot15d4\x10\x80\x80\x80 \x12\x10\n\tSixLowPan\x10\x80\x80\x80(\x12\n\n\x03\x45sb\x10\x80\x80\x80\x30\x12\x17\n\x10LogitechUnifying\x10\x80\x80\x80\x38\x12\r\n\x06Mosart\x10\x80\x80\x80@\x12\n\n\x03\x41NT\x10\x80\x80\x80H\x12\x0f\n\x08\x41NT_Plus\x10\x80\x80\x80P\x12\r\n\x06\x41NT_FS\x10\x80\x80\x80X*P\n\nDeviceType\x12\x12\n\x0e\x45sp32BleFuzzer\x10\x00\x12\r\n\tButterfly\x10\x01\x12\x0c\n\x08\x42tleJack\x10\x02\x12\x11\n\rVirtualDevice\x10\x04*|\n\nCapability\x12\x0c\n\x08_CapNone\x10\x00\x12\x08\n\x04Scan\x10\x01\x12\t\n\x05Sniff\x10\x02\x12\n\n\x06Inject\x10\x04\x12\x07\n\x03Jam\x10\x08\x12\n\n\x06Hijack\x10\x10\x12\x08\n\x04Hook\x10 \x12\x10\n\x0cSimulateRole\x10@\x12\x0e\n\tNoRawData\x10\x80\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'whad.protocol.device_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _DEVICEINFORESP.fields_by_name['capabilities']._options = None
  _DEVICEINFORESP.fields_by_name['capabilities']._serialized_options = b'\020\001'
  _DOMAIN._serialized_start=875
  _DOMAIN._serialized_end=1073
  _DEVICETYPE._serialized_start=1075
  _DEVICETYPE._serialized_end=1155
  _CAPABILITY._serialized_start=1157
  _CAPABILITY._serialized_end=1281
  _DEVICERESETQUERY._serialized_start=41
  _DEVICERESETQUERY._serialized_end=59
  _DEVICEREADYRESP._serialized_start=61
  _DEVICEREADYRESP._serialized_end=78
  _SETTRANSPORTSPEED._serialized_start=80
  _SETTRANSPORTSPEED._serialized_end=114
  _DEVICEINFORESP._serialized_start=117
  _DEVICEINFORESP._serialized_end=341
  _DEVICEDOMAININFORESP._serialized_start=343
  _DEVICEDOMAININFORESP._serialized_end=409
  _DEVICEINFOQUERY._serialized_start=411
  _DEVICEINFOQUERY._serialized_end=447
  _DEVICEDOMAININFOQUERY._serialized_start=449
  _DEVICEDOMAININFOQUERY._serialized_end=488
  _MESSAGE._serialized_start=491
  _MESSAGE._serialized_end=872
# @@protoc_insertion_point(module_scope)



from whad.ble.stack.smp.exceptions import SMInvalidCustomFunction
from whad.ble.stack.smp.constants import IOCAP_KEYBD_ONLY, IOCAP_DISPLAY_ONLY, \
    IOCAP_DISPLAY_YESNO, IOCAP_KEYBD_DISPLAY, IOCAP_NOINPUT_NOOUTPUT

class Pairing:
    def __init__(
        self,
        oob=False,
        bonding=True,
        mitm=False,
        lesc=True,
        keypress=False,
        ct2=False,
        max_key_size=16,
        iocap=IOCAP_NOINPUT_NOOUTPUT,
        enc_key=True,
        id_key=True,
        sign_key=True,
        link_key=True,
        accept_pairing=True,
        custom_functions=None
    ):
        self.oob = oob
        self.bonding = bonding
        self.mitm = mitm
        self.lesc = lesc
        self.keypress = keypress
        self.ct2 = ct2
        self.max_key_size = max_key_size
        self.iocap = iocap
        self.enc_key = enc_key
        self.id_key = id_key
        self.sign_key = sign_key
        self.link_key = link_key
        self.accept_pairing = accept_pairing
        self.custom_functions = custom_functions


class PairingCustomFunctions:
    def __init__(self,
            check_initiator_legacy_confirm = None,
            check_responder_legacy_confirm = None,
            compute_exchange_value = None,
            compute_ltk_and_mackey = None,
            compute_lesc_numeric_comparison = None,
            check_lesc_confirm_value = None,
            compute_lesc_confirm_value = None,
            compute_legacy_confirm_value = None,
            check_lesc_numeric_comparison = None,
            generate_p256_keypair = None,
            generate_legacy_ltk = None,
            generate_legacy_random = None,
            generate_legacy_ediv = None,
            generate_irk = None,
            generate_csrk = None,
            get_passkey_entry = None,
            get_pin_code = None
    ):
        self.check_initiator_legacy_confirm = check_initiator_legacy_confirm
        self.check_responder_legacy_confirm = check_responder_legacy_confirm
        self.compute_exchange_value = compute_exchange_value
        self.compute_ltk_and_mackey = compute_ltk_and_mackey
        self.compute_lesc_numeric_comparison = compute_lesc_numeric_comparison
        self.check_lesc_confirm_value = check_lesc_confirm_value
        self.compute_lesc_confirm_value = compute_lesc_confirm_value
        self.compute_legacy_confirm_value = compute_legacy_confirm_value
        self.check_lesc_numeric_comparison = check_lesc_numeric_comparison
        self.generate_p256_keypair = generate_p256_keypair
        self.generate_legacy_ltk = generate_legacy_ltk
        self.generate_legacy_random = generate_legacy_random
        self.generate_legacy_ediv = generate_legacy_ediv
        self.generate_irk = generate_irk
        self.generate_csrk = generate_csrk
        self.get_passkey_entry = get_passkey_entry
        self.get_pin_code = get_pin_code



"""
HCI/WHAD messages converter.
"""
import logging

from struct import unpack
from queue import Queue
from enum import IntEnum
from typing import List

# Scapy
from scapy.layers.bluetooth import HCI_Event_LE_Meta, HCI_LE_Meta_Advertising_Reports, \
    HCI_LE_Meta_Connection_Complete, L2CAP_Hdr, HCI_Hdr, HCI_ACL_Hdr, \
    HCI_Event_Disconnection_Complete, HCI_LE_Meta_Long_Term_Key_Request, \
    HCI_Event_Encryption_Change
from whad.scapy.layers.bluetooth import HCI_LE_Meta_Data_Length_Change
from scapy.layers.bluetooth4LE import BTLE_DATA, BTLE_CTRL, LL_ENC_REQ, \
    LL_ENC_RSP, LL_START_ENC_RSP, LL_START_ENC_REQ
from scapy.compat import raw

# Whad
from whad.scapy.layers.hci import HCI_LE_Meta_Enhanced_Connection_Complete
from whad.exceptions import WhadDeviceUnsupportedOperation
from whad.hub.ble import BleAddrType, BleDirection, BDAddress, AdvType, HubMessage

logger = logging.getLogger(__name__)

class HCIRole(IntEnum):
    """HCI role.
    """
    NONE = 0
    CENTRAL = 1
    PERIPHERAL = 2

class HCIConverter:
    """
    This class converts HCI event to WHAD BLE packets and WHAD BLE packets to HCI commands.
    It allows to harmonise the representation of Bluetooth Low Energy traffic, as WHAD expects
    Link Layer-like packets.

    Obviously, some packets can't be converted easily from a representation to the other one without
    losing information or making asumptions. As a result, you should expect some limitations.
    """

    def __init__(self, device):
        self.__device = device
        self.role = HCIRole.NONE
        self.pending_messages_queue = Queue()
        self.pending_key_request = False
        self.cached_l2cap_payload = b""
        self.cached_l2cap_length = 0
        self.waiting_l2cap_fragments = False
        self.__locked = False

    def lock(self):
        """Lock the message converter. Messages are then added in a list
        of pending messages and will be processed when converter is unlocked.
        """
        self.__locked = True

    def unlock(self):
        """Unlock converter and trigger pending messages processing.
        """
        self.__locked = False

    def split_l2cap_into_hci(self, l2cap_data: bytes, conn_handle: int):
        """Split L2CAP packet into multiple HCI packets.
        """
        max_acl_len = self.__device.get_max_acl_len()
        nb_hci_packets = len(l2cap_data)//max_acl_len
        if len(l2cap_data)%max_acl_len > 0:
            nb_hci_packets += 1
        hci_packets = []
        for i in range(nb_hci_packets):
            pkt = HCI_Hdr() / HCI_ACL_Hdr(handle = conn_handle, PB=1 if i>0 else 0)
            pkt = pkt / l2cap_data[i*max_acl_len:(i+1)*max_acl_len]
            hci_packets.append(pkt)
        logger.debug("[hci converter] split ACL data into %d chunks (total: %d, acl_len: %d)", nb_hci_packets, len(l2cap_data), max_acl_len)
        return hci_packets

    def process_message(self, message: HubMessage):
        """This function turns a BLE hub message into the corresponding HCI
        packet.

        :param message: Hub message to convert.
        :type message: HubMessage
        """
        ll_packet = BTLE_DATA(message.pdu)

        # L2CAP packet reassembly
        if L2CAP_Hdr in ll_packet or self.waiting_l2cap_fragments:
            if not self.waiting_l2cap_fragments and (
                    (len(raw(ll_packet[L2CAP_Hdr:])) - 4) < ll_packet[L2CAP_Hdr:].len):
                self.waiting_l2cap_fragments = True
                self.cached_l2cap_payload = raw(ll_packet[BTLE_DATA:][1:])
                self.cached_l2cap_length = ll_packet[L2CAP_Hdr:].len

                # No HCI packet to send for now (data queued)
                logger.debug("l2cap is incomplete (%s/%s)",
                             len(self.cached_l2cap_payload)-4,
                             self.cached_l2cap_length)
                return []

            if self.waiting_l2cap_fragments:
                self.cached_l2cap_payload += raw(ll_packet[BTLE_DATA:][1:])
                if self.cached_l2cap_length <= (len(self.cached_l2cap_payload) - 4):
                    if self.cached_l2cap_length < len(self.cached_l2cap_payload) - 4:
                        logger.debug("[hci device][%s] too much data (got %d, expected %d)",
                                     self.__device.interface, len(self.cached_l2cap_payload) - 4,
                                     self.cached_l2cap_length)
                    self.waiting_l2cap_fragments = False
                    logger.debug("[hci device] reassembled l2cap data !")

                    # L2CAP data has been reassembled, then split in respect of
                    # the underlying device max ACL length
                    return self.split_l2cap_into_hci(
                        bytes(L2CAP_Hdr(self.cached_l2cap_payload)),
                        message.conn_handle
                    )

                # No HCI packet for now.
                logger.debug("l2cap is incomplete, more fragments needed (%s/%s).", len(self.cached_l2cap_payload) - 4, self.cached_l2cap_length)
                return []

            # L2CAP packet is complete but must be split
            l2cap_data = bytes(ll_packet[L2CAP_Hdr:])
            if len(l2cap_data) > self.__device.get_max_acl_len():
                return self.split_l2cap_into_hci(l2cap_data, message.conn_handle)

            # No fragmentation, send data as-is.
            hci_packet = HCI_Hdr() / HCI_ACL_Hdr(handle = message.conn_handle)
            hci_packet = hci_packet / ll_packet[L2CAP_Hdr:]
            logger.debug("l2cap does not need frag")
            return [hci_packet]

        ll_packet.show()
        if ll_packet.LLID == 3:
            if LL_ENC_REQ in ll_packet:
                # When a LL_ENC_REQ is received from BLE stack, we issue a fake LL_ENC_RESP
                # with `skds` and `ivs` set to 0. 
                pdu = BTLE_DATA() / BTLE_CTRL() / LL_ENC_RSP(skds=0, ivs=0)

                direction = (BleDirection.SLAVE_TO_MASTER if
                             self.role == HCIRole.CENTRAL else
                             BleDirection.MASTER_TO_SLAVE
                )
                processed = False
                conn_handle = message.conn_handle

                msg = self.__device.hub.ble.create_pdu_received(
                    direction,
                    raw(pdu),
                    conn_handle,
                    processed
                )

                self.add_pending_message(msg)

                # We also issue a fake LL_START_ENC_REQ to our stack like a peripheral
                # would do to enable encryption at the link layer. We should receive a
                # LL_START_ENC_RSP in response.
                pdu = BTLE_DATA() / BTLE_CTRL() / LL_START_ENC_REQ()

                direction = (BleDirection.SLAVE_TO_MASTER if
                             self.role == HCIRole.CENTRAL else
                             BleDirection.MASTER_TO_SLAVE
                )
                processed = False
                conn_handle = message.conn_handle

                msg = self.__device.hub.ble.create_pdu_received(
                    direction,
                    raw(pdu),
                    conn_handle,
                    processed
                )

                self.add_pending_message(msg)

                return []

            #logger.warning("HCI devices cannot send control PDU.")
            raise WhadDeviceUnsupportedOperation(
                "send_pdu", "Device cannot send control PDU, only data PDU."
            )

        # No HCI packet to send
        return []

    def add_pending_message(self, event):
        """Add pending message to queue.

        :param event: Event message to add to queue
        :type event: HubMessage
        """
        self.pending_messages_queue.put(event)

    def get_pending_messages(self) -> List[HubMessage]:
        """Get pending messages.

        :return: List of pending hub messages
        :rtype: list
        """
        events = []
        while not self.pending_messages_queue.empty():
            events.append(self.pending_messages_queue.get())
        return events

    def process_event(self, event):
        """
        Converts an HCI event to the corresponding Link Layer packet (if possible).
        """

        if HCI_Event_LE_Meta in event:
            if HCI_LE_Meta_Advertising_Reports in event:
                return self.process_advertising_reports(event[HCI_LE_Meta_Advertising_Reports:])

            if HCI_LE_Meta_Connection_Complete in event:
                return self.process_connection_complete(event[HCI_LE_Meta_Connection_Complete:])

            if HCI_LE_Meta_Enhanced_Connection_Complete in event:
                return self.process_connection_complete(
                    event[HCI_LE_Meta_Enhanced_Connection_Complete:]
                )

            if HCI_LE_Meta_Long_Term_Key_Request in event:
                return self.process_long_term_key_request(event[HCI_LE_Meta_Long_Term_Key_Request:])
            
            if HCI_LE_Meta_Data_Length_Change in event:
                # Process a data length change
                length = event[HCI_LE_Meta_Data_Length_Change].max_tx_octets
                logger.debug("[hci][%s] update HCI data length to %d", self.__device.interface, length)
                self.__device._update_max_acl_len(length)


        if HCI_Event_Encryption_Change in event:
            return self.process_encryption_change_event(event[HCI_Event_Encryption_Change:])

        if HCI_ACL_Hdr in event:
            return self.process_acl_data(event[HCI_ACL_Hdr:])

        if HCI_Event_Disconnection_Complete in event:
            return self.process_disconnection_complete(event[HCI_Event_Disconnection_Complete:])

        # Return no messages on unknown events
        return []

    def process_encryption_change_event(self, event) -> List[HubMessage]:
        """Process encryption change.

        :param event: HCI event to process
        :return: list of hub messages
        :rtype: list
        """
        pdu = BTLE_DATA()/BTLE_CTRL()/LL_START_ENC_RSP()

        direction = (BleDirection.SLAVE_TO_MASTER if
                     self.role == HCIRole.CENTRAL else
                     BleDirection.MASTER_TO_SLAVE
        )
        processed = False
        conn_handle = event.handle

        msg = self.__device.hub.ble.create_pdu_received(
            direction,
            raw(pdu),
            conn_handle,
            processed
        )

        return [msg]

    def process_long_term_key_request(self, event):
        """Process long-term key request event.
        """
        pdu = BTLE_DATA()/BTLE_CTRL()/LL_ENC_REQ(
            rand = unpack('<Q', event.rand)[0],
            ediv = event.ediv,
            # SKD and IV are processed in Link Layer and are not transmitted to Host,
            # so we can't provide the real value. We arbitrarily set them to 0 to allow
            # BLE stack to continue its operations.
            skdm = 0,
            ivm = 0
        )


        direction = (BleDirection.SLAVE_TO_MASTER if
                     self.role == HCIRole.CENTRAL else
                     BleDirection.MASTER_TO_SLAVE
        )
        processed = False
        conn_handle = event.handle

        msg = self.__device.hub.ble.create_pdu_received(
            direction,
            raw(pdu),
            conn_handle,
            processed
        )

        self.pending_key_request = True
        return [msg]

    def process_acl_data(self, event):
        """Process ACL data event.
        """
        if event.PB == 2 and L2CAP_Hdr in event:
            pdu = BTLE_DATA()/event[L2CAP_Hdr:]

            # If HCI ACL Data PB flag==1 then it is a continued fragment.
            # We make sure BTLE_DATA.LLID is then 0x01 (Continuation
            # of a L2CAP message), our stack will take care of packet reassembly.
            if event.PB == 0x01:
                pdu.LLID = 1

            direction = (BleDirection.SLAVE_TO_MASTER if
                         self.role == HCIRole.CENTRAL else
                         BleDirection.MASTER_TO_SLAVE
            )
            processed = False
            conn_handle = event.handle

            logger.debug("[hci] received PDU from remote device, forwarding to host")
            msg = self.__device.hub.ble.create_pdu_received(
                direction,
                raw(pdu),
                conn_handle,
                processed
            )

            if self.__locked:
                self.add_pending_message(msg)
                return []
            else:
                return [msg]

        if event.PB == 1:
            pdu = BTLE_DATA()/event[HCI_ACL_Hdr:].payload

            # If HCI ACL Data PB flag==1 then it is a continued fragment.
            # We make sure BTLE_DATA.LLID is then 0x01 (Continuation
            # of a L2CAP message), our stack will take care of packet reassembly.
            pdu.LLID = 1

            direction = (BleDirection.SLAVE_TO_MASTER if
                         self.role == HCIRole.CENTRAL else
                         BleDirection.MASTER_TO_SLAVE
            )
            processed = False
            conn_handle = event.handle

            msg = self.__device.hub.ble.create_pdu_received(
                direction,
                raw(pdu),
                conn_handle,
                processed
            )
            if self.__locked:
                self.add_pending_message(msg)
                return []
            else:
                return [msg]

        # Unsupported event, no messages.
        return []

    def process_connection_complete(self, event):
        """Process connection complete HCI event.
        """
        if event.status == 0x00:

            # Mark device as connected and register handle.
            self.__device.on_connection_created(event.handle)

            # Send BLE connected message to consumer.
            handle = event.handle
            if event.role == 0: # master role
                self.role = HCIRole.CENTRAL
                initiator_address = self.__device._bd_address.value
                initiator_address_type = self.__device._bd_address.type
                responder_address = bytes.fromhex(event.paddr.replace(":",""))[::-1]
                responder_address_type = (
                    BleAddrType.PUBLIC if event.patype == 0 else BleAddrType.RANDOM
                )
            else:
                self.role = HCIRole.PERIPHERAL
                initiator_address = bytes.fromhex(event.paddr.replace(":",""))[::-1]
                initiator_address_type = (
                    BleAddrType.PUBLIC if event.patype == 0 else BleAddrType.RANDOM
                )
                responder_address = self.__device._bd_address.value
                responder_address_type = self.__device._bd_address.type

            msg = self.__device.hub.ble.create_connected(
                BDAddress(initiator_address, addr_type=initiator_address_type),
                BDAddress(responder_address, addr_type=responder_address_type),
                0, # No access address
                handle
            )

            return [msg]
        else:
            logger.debug("[%s] Connection complete event with status 0x%x", self.__device.interface,
                         event.status)
        # Nothing to convert
        return []

    def process_disconnection_complete(self, event):
        """Process HCI disconnection event.
        """
        if event.status == 0x00:
            # Mark device as disconnected and remove handle
            #self.__device._connected = False
            #self.__device._active_handles.remove(event.handle)
            self.__device.on_connection_terminated(event.handle)

            # Send disconnection message to consumer.
            logger.debug("[hci] sending Disconnected message to host")
            msg = self.__device.hub.ble.create_disconnected(
                event.reason,
                event.handle
            )

            return [msg]

    def process_advertising_reports(self, reports):
        """Process HCI advertising reports.
        """
        messages = []
        for report in reports.reports:

            adv_type = AdvType.ADV_UNKNOWN
            if report.type == 0:
                adv_type = AdvType.ADV_IND
            elif report.type == 1:
                adv_type = AdvType.ADV_DIRECT_IND
            elif report.type == 2:
                adv_type = AdvType.ADV_SCAN_IND
            elif report.type == 3:
                adv_type = AdvType.ADV_NONCONN_IND
            elif report.type == 4:
                adv_type = AdvType.ADV_SCAN_RSP

            # Flatten EIR data
            eir_data = b""
            for data in report.data:
                eir_data += raw(data)

            msg = self.__device.hub.ble.create_adv_pdu_received(
                adv_type,
                report.rssi if hasattr(report, "rssi") else 0,
                BDAddress(report.addr, random = not report.atype == 0),
                bytes(eir_data)
            )

            messages.append(msg)
        return messages



"""Import all known layers by default

This can be useful if you plan to read or write PCAP packets as DLT are also
declared in these submodules.
"""
from .phy import *
from .dot15d4tap import *
from .nordic import *
from .bluetooth import *
from .lorawan import *
from .rf4ce import *
from .esb import *
from .unifying import *
from .zdp import *
from .zll import *

